import timeit
import warnings
from functools import partial
from itertools import product

import numpy as np
try:  # only necessary if multiple cores should be used
    from multiprocessing_on_dill.pool import Pool as DillPool
except ModuleNotFoundError:
    pass
from datetime import datetime
import scipy.optimize as sciopt
from scipy.optimize._constraints import old_bound_to_new  # noqa
from scipy.special import expit
from numdifftools import Hessian

from .util import TAB, SP2
from .util import _slsqp_epsilon

def loop(fun, params, args, gridsize, minimize_along_grid, bounds, verbosity, param_id):
    # t0 = timeit.default_timer()
    if param_id and (verbosity > 1) and (np.mod(param_id, 1000) == 0):
        print(f'{TAB}{TAB}Grid iteration {param_id} / {gridsize}')
    negll = fun(params[param_id], *args)

    if minimize_along_grid:
        fit = sciopt.minimize(fun, params[param_id], bounds=bounds, args=tuple(args), method='slsqp')
        x = fit.x
        negll = fit.fun
    else:
        x = params[param_id]

    # print(timeit.default_timer() - t0)
    return negll, x


def fgrid(fun, valid, args, num_cores=1, minimize_along_grid=False, bounds=None, verbosity=1):
    if num_cores > 1:
        with DillPool(num_cores) as pool:
            result = pool.map(partial(loop, fun, valid, args, len(valid), minimize_along_grid, bounds, verbosity), range(len(valid)))
            negll_grid, x_grid = [res[0] for res in result], [res[1] for res in result]
    else:
        negll_grid = [None] * len(valid)
        x_grid = [None] * len(valid)
        for i, param in enumerate(valid):
            negll_grid[i], x_grid[i] = loop(fun, valid, args, len(valid), minimize_along_grid, bounds, verbosity, i)
    return negll_grid, x_grid


# def fine_grid_search(x0, fun, args, param_set, ll_grid, valid, gridsearch_resolution, n_grid_iter, multiproc=False,
#                      multiproc_cores=1, verbosity=1):
#     n_grid_candidates_init = 3 * gridsearch_resolution
#     gx0 = x0
#     gll_min_grid = np.min(ll_grid)
#     grid_candidates = [valid[i] for i in np.argsort(ll_grid)[:n_grid_candidates_init]]
#     previous_grid_range = [param_set.grid_range for _ in range(n_grid_candidates_init)]
#     candidate_ids = list(range(n_grid_candidates_init))
#     counter = 0
#     for i in range(n_grid_iter):
#         if (verbosity > 1) and (np.mod(i, 10) == 0):
#             print(f'{TAB}Grid iteration {i + 1} / {n_grid_iter}')
#         gvalid = []
#         valid_candidate_ids = []
#         grid_range = [None] * len(grid_candidates) * 2
#         for j, grid_candidate in enumerate(grid_candidates):
#             grid_range[j] = [None] * param_set.nparams_flat
#             for k, p in enumerate(grid_candidate):
#                 ind = np.where(previous_grid_range[candidate_ids[j]][k] == p)[0][0]
#                 lb = previous_grid_range[candidate_ids[j]][k][max(0, ind - 1)]  # noqa
#                 ub = previous_grid_range[candidate_ids[j]][k][  # noqa
#                     min(len(previous_grid_range[candidate_ids[j]][k]) - 1, ind + 1)]  # noqa
#                 grid_range[j][k] = np.around(np.linspace(lb, ub, 5), decimals=10)  # noqa
#             if param_set.constraints is None:
#                 gvalid_candidate = list(product(*grid_range[j]))
#             else:
#                 gvalid_candidate = [p for p in product(*grid_range[j]) if  # noqa
#                                     np.all([con['fun'](p) >= 0 for con in param_set.constraints])]
#             gvalid += gvalid_candidate
#             valid_candidate_ids += [j] * len(gvalid_candidate)
#             if verbosity > 1:
#                 print(f'{TAB}{TAB}Candidate {j + 1}: {[(p[0], p[-1]) for p in grid_range[j]]}')  # noqa
#         gll_grid = fgrid(fun, gvalid, args, multiproc, multiproc_cores, verbosity=0)[0]
#         counter += len(gvalid)
#
#         min_id = np.argmin(gll_grid)
#         if gll_grid[min_id] < gll_min_grid:
#             gll_min_grid = gll_grid[min_id]
#             gx0 = gvalid[min_id]
#
#         if verbosity:
#             print(f'{TAB}{TAB}Best fit: {gx0} (LL={gll_min_grid})')
#
#         if i != n_grid_iter - 1:
#             grid_candidates, candidate_ids = [], []
#             for j in np.argsort(gll_grid):
#                 if gvalid[j] not in grid_candidates:
#                     grid_candidates += [gvalid[j]]
#                     candidate_ids += [valid_candidate_ids[j]]
#                     if len(candidate_ids) == gridsearch_resolution:
#                         break
#             previous_grid_range = grid_range
#     fit = sciopt.OptimizeResult(success=True, x=gx0, fun=gll_min_grid, nfev=counter)
#
#     return fit


def subject_estimation(fun, param_set, args=(), gridsearch=False, num_cores=1,
                       minimize_along_grid=False, global_minimization=None,
                       # n_grid_candidates=10, n_grid_iter=3,
                       scipy_solvers='slsqp', slsqp_epsilon=_slsqp_epsilon,
                       verbosity=1, silence_warnings=False):

    t0 = timeit.default_timer()

    # if verbosity:
    #     negll_initial_guess = fun(param_set.guess, *args)
    #     print(f'{TAB}{TAB}Initial guess (neg. LL: {negll_initial_guess:.2f})')
    #     for i, p in enumerate(param_set.param_names_flat):
    #         if p.startswith('type2_criteria') and not p.endswith('0'):
    #             criterion_id = int(p.split('_')[-1])
    #             criterion = param_set.guess[i-criterion_id:i+1].sum()
    #             print(f'{TAB}{TAB}{TAB}[guess] {p}: {param_set.guess[i]:.4g} = gap | criterion = {criterion:.4g}')
    #         else:
    #             print(f'{TAB}{TAB}{TAB}[guess] {p}: {param_set.guess[i]:.4g}')

    bounds = sciopt.Bounds(*old_bound_to_new(param_set.bounds), keep_feasible=True)

    x0_guess = param_set.guess
    fit_guess = sciopt.OptimizeResult(success=True, x=x0_guess, fun=fun(x0_guess, *args), nfev=1)
    if gridsearch:
        if param_set.constraints is not None and len(param_set.constraints):
            valid = np.array([p for p in product(*param_set.grid_range) if
                     np.all([con['fun'](p) >= 0 for con in param_set.constraints])])
        else:
            valid = np.array(list(product(*param_set.grid_range)))
        if verbosity > 1:
            print(f"{TAB}{TAB}Grid search activated (grid size = {len(valid)})")
        t0 = timeit.default_timer()
        ll_grid, x_grid = fgrid(fun, valid, args, num_cores, minimize_along_grid, bounds, verbosity=verbosity)
        x0_grid = x_grid[np.argmin(ll_grid)]
        ll_min_grid = np.min(ll_grid)
        grid_time = timeit.default_timer() - t0
        if verbosity > 1:
            for i, p in enumerate(param_set.param_names_flat):
                if p.startswith('type2_criteria') and not p.endswith('0'):
                    criterion_id = int(p.split('_')[-1])
                    criterion = param_set.guess[i-criterion_id:i+1].sum()
                    print(f'{TAB}{TAB}{TAB}[grid] {p}: {param_set.guess[i]:.4g} = gap | criterion = {criterion:.4g}')
                else:
                    print(f'{TAB}{TAB}{TAB}[grid] {p}: {x0_grid[i]:.4g}')
            print(f"{TAB}{TAB}Grid neg. LL: {ll_min_grid:.1f}")
            print(f"{TAB}{TAB}Grid runtime: {grid_time:.2f} secs")
        fit_grid = sciopt.OptimizeResult(success=True, x=x0_grid, fun=ll_min_grid, nfev=len(valid))
        fit_best = fit_grid if ll_min_grid < fit_guess.fun else fit_guess
    else:
        fit_best = fit_guess
        x0_grid = None

    # if fine_gridsearch:
    #     fit_grid = fine_grid_search(x0_grid if gridsearch else x0_guess, fun, args, param_set, ll_grid, valid, n_grid_candidates,
    #                                 n_grid_iter, multiproc, multiproc_cores, verbosity=verbosity)
    #     x0_grid = fit_grid.x
    #     fit_best = fit_grid if fit_grid.fun < fit_guess.fun else fit_guess

    if global_minimization is not None:
        # if verbosity:
        #     print(f'{TAB}{TAB}Performing MLE (global minimization)')
        t_global = timeit.default_timer()
        # fit_global = basinhopping(fun, fit_best.x, take_step=RandomDisplacementBoundsConstraints(bounds, param_set.constraints),
        #                           accept_test=BoundsConstraints(bounds, param_set.constraints),
        #                           minimizer_kwargs=dict(method='Nelder-Mead', args=tuple(args)))
        if global_minimization == 'shgo':
            fit_global = sciopt.shgo(fun, bounds=bounds, args=tuple(args), minimizer_kwargs=dict(method='slsqp', bounds=bounds),
                                     options=dict(sampling_method='sobol', maxiter=50))
        elif global_minimization == 'differential_evolution':
            try:
                fit_global = sciopt.differential_evolution(fun, bounds=bounds, args=tuple(args))
            except ValueError as e:
                fit_global = sciopt.shgo(fun, bounds=bounds, args=tuple(args), minimizer_kwargs=dict(method='slsqp', bounds=bounds),
                                     options=dict(sampling_method='sobol', maxiter=50))
                print(e)

        elif global_minimization == 'dual_annealing':
            fit_global = sciopt.dual_annealing(fun, bounds=bounds, args=tuple(args),
                                               minimizer_kwargs={"method": "slsqp", "bounds": bounds})
        else:
            raise ValueError(f"Unknown global minimization {global_minimization}. Please choose one of 'shgo', "
                             f"'dual_annealing' or 'differential_evolution'.")

        execution_time_global = timeit.default_timer() - t_global
        x0_global = fit_global.x
        if verbosity:
            print(f'{TAB}{TAB}{TAB}.. global MLE finished ({execution_time_global:.1f} secs).')
    else:
        x0_global = None
        execution_time_global = None

    # if verbosity:
    #     print(f'{TAB}{TAB}Performing MLE (local minimization)')

    t_local = timeit.default_timer()

    scipy_solvers = [scipy_solvers] if isinstance(scipy_solvers, str) else scipy_solvers

    # We start both from x0_guess and x0_grid, as x0_grid is more prone to local minima
    x0s = (x0_guess, x0_grid, x0_global)

    best_solver = f"init_{'guess' if x0_grid is None else 'grid'}" if global_minimization is None else global_minimization
    for i, x0 in enumerate(x0s):
        if x0 is not None:
            for method in scipy_solvers:
                if method == 'slsqp':
                    slsqp_epsilon_ = slsqp_epsilon if hasattr(slsqp_epsilon, '__len__') else [slsqp_epsilon]
                    for eps in slsqp_epsilon_:
                        fit = sciopt.minimize(fun, x0, bounds=bounds, args=tuple(args), constraints=param_set.constraints,
                                              method='slsqp', options=dict(eps=eps))
                        if fit.fun < fit_best.fun:
                            fit_best = fit
                            best_solver = f"slsqp_{eps:.3g}_init{('guess', 'grid', 'global')[i]}"
                else:
                    try:
                        fit = sciopt.minimize(fun, x0, bounds=bounds, args=tuple(args), constraints=param_set.constraints,
                                              method=method)
                    except ValueError as e:
                        if 'violates bound constraints' in str(e) and (method == 'trust-constr') and check_x0_within_bounds(x0, bounds):
                            # Catch this error in trust-constr which is a bug in scipy that has been fixed soon after
                            # the release of scipy 1.17.0:
                            # https://github.com/scipy/scipy/pull/24454
                            if not silence_warnings:
                                warnings.warn('SciPy trust-constr evaluated outside bounds despite a feasible x0. '
                                              'Using L-BFGS-B instead.')
                            method = 'L-BFGS-B'
                            fit = sciopt.minimize(fun, x0, bounds=bounds, args=tuple(args),
                                                  constraints=param_set.constraints, method=method)
                        else:
                            raise

                    if fit.fun < fit_best.fun:
                        fit_best = fit
                        best_solver = f"{method}_init{('guess', 'grid', 'global')[i]}"
    fit_best.execution_time = timeit.default_timer() - t0  # noqa
    fit_best.execution_time_global = execution_time_global
    fit_best.execution_time_local = timeit.default_timer() - t_local
    fit_best.best_solver = best_solver
    fit_best.x0_grid = x0_grid
    # if verbosity:
    #     print(f'{TAB}{TAB}{TAB}.. local MLE finished ({fit_best.execution_time_local:.1f} secs).')


    H = Hessian(fun)(fit_best.x)
    H = 0.5 * (H + H.T)  # make sure it is not asymmetric due to numerical errors
    # Logic: if not singular and all_eigenvalues > 0
    # if not (singular := (np.linalg.cond(H) > 1 / np.finfo(H.dtype).eps) and np.all(np.diag(np.linalg.inv(H)) > 0):
    # if not (singular := (np.linalg.svd(H, compute_uv=False)[-1] < np.finfo(H.dtype).eps)) and np.all(np.diag(np.linalg.inv(H)) > 0):
    if not (singular := np.linalg.cond(H) > 1e12) and np.all(np.diag(np.linalg.inv(H)) > 0):
        # positive definite
        fit_best.hessian = H
    else:
        if not silence_warnings:
            warnings.warn(f'Hessian is {"singular" if singular else "not positive definite"}. Standard errors are computed '
                          f'with a ridged Hessian and may be unreliable for some parameters (check extreme values).')
        fit_best.hessian = ridge_hessian(H)

    return fit_best


# def ridge_hessian_se(H, eps_scale=1e-12, max_iter=60):
#     # Apply ridge regression until the Hessian H is positive definite
#     Hs = 0.5 * (H + H.T)
#     n = Hs.shape[0]
#
#     scale = max(np.linalg.norm(Hs, ord=np.inf), 1.0)
#     lam = eps_scale * scale
#     for _ in range(max_iter):
#         Hr = Hs + lam * np.eye(n)
#         try:
#             L = np.linalg.cholesky(Hr)
#             break
#         except np.linalg.LinAlgError:
#             lam *= 10
#     else:
#         warnings.warn('Could not determine uncertainty of the parameter estimate, even with a ridged Hessian.')
#
#     I = np.eye(n)
#     Y = np.linalg.solve(L, I)
#     X = np.linalg.solve(L.T, Y)
#     se = np.sqrt(np.diag(X))
#     # cond = np.linalg.cond(Hr)
#     # return se, Hr, lam, cond
#     return se


def ridge_hessian(H, eps_scale=1e-12, max_iter=60):
    # Apply ridge regression until the Hessian H is positive definite
    Hs = 0.5 * (H + H.T)
    n = Hs.shape[0]

    scale = max(np.linalg.norm(Hs, ord=np.inf), 1.0)
    lam = eps_scale * scale
    for _ in range(max_iter):
        Hr = Hs + lam * np.eye(n)
        try:
            np.linalg.cholesky(Hr)
            break
        except np.linalg.LinAlgError:
            lam *= 10
    else:
        warnings.warn('Could not determine uncertainty of the parameter estimate, even with a ridged Hessian.')
        return None

    return Hr


def check_x0_within_bounds(x0, bounds, tol=0.0):

    x0 = np.asarray(x0, dtype=float)

    if hasattr(bounds, 'lb'):  # scipy.optimize.Bounds
        lb = np.asarray(bounds.lb, dtype=float)
        ub = np.asarray(bounds.ub, dtype=float)
    else:  # list of (lb, ub)
        lb = np.array([b[0] for b in bounds], dtype=float)
        ub = np.array([b[1] for b in bounds], dtype=float)

    bad = (
        (x0 < lb - tol) |
        (x0 > ub + tol) |
        ~np.isfinite(x0)
    )

    return not np.any(bad)


def laplace_cov_from_hessian(H, eig_floor=1e-10):
    """
    Robust inverse of a small symmetric PD-ish Hessian.
    Returns covariance approx Sigma = H^{-1}.
    """
    H = 0.5 * (H + H.T)  # symmetrize

    # Cholesky path (fast + stable when PD)
    try:
        L = np.linalg.cholesky(H)
        I = np.eye(H.shape[0])
        Linv = np.linalg.solve(L, I)
        return Linv.T @ Linv
    except np.linalg.LinAlgError:
        # Eigenvalue-floor fallback
        w, Q = np.linalg.eigh(H)
        w = np.maximum(w, eig_floor)
        return (Q * (1.0 / w)) @ Q.T


def d_transform_constrained_du(u, bounds):
    """
    Elementwise derivative d(params)/d(uparams) for your transform_to_constrained_space.
    u: (..., K) or (K,)
    bounds: (K, 2) with columns [lower, upper]
    returns: same shape as u
    """
    lower, upper = bounds.T
    u = np.asarray(u, dtype=float)

    # broadcast masks across last axis
    m_unbounded  = np.isinf(lower) & np.isinf(upper)
    m_lower_only = np.isfinite(lower) & np.isinf(upper)
    m_upper_only = np.isinf(lower) & np.isfinite(upper)
    m_bounded    = np.isfinite(lower) & np.isfinite(upper)

    deriv = np.empty_like(u, dtype=float)

    deriv[..., m_unbounded] = 1.0

    # Match your exp clipping (avoid overflow)
    z = np.clip(u, -709.0, 709.0)
    ez = np.exp(z)

    deriv[..., m_lower_only] = ez[..., m_lower_only]          # d(l + exp(u))/du = exp(u)
    deriv[..., m_upper_only] = -ez[..., m_upper_only]         # d(u - exp(u))/du = -exp(u)

    s = expit(u[..., m_bounded])
    deriv[..., m_bounded] = (upper[m_bounded] - lower[m_bounded]) * s * (1.0 - s)

    return deriv


def compute_fixed_effect_se(fun, uparams_init, uparams_fe, bounds, idx_fe, eps=1e-4, eps_scale=True):
    """
    Compute global FE SEs (in constrained params-space) using Hessian of the group FE objective,
    conditional on current uparams_init.

    Returns:
      fe_se: fixed effects standard errors
    """

    def obj_fe_only(x_packed):
        u_tmp = uparams_fe.copy()
        u_tmp[idx_fe] = x_packed
        return objective_group_fe(u_tmp, fun, uparams_init, bounds, idx_fe)

    xhat_u = uparams_fe[idx_fe].copy()
    H = numerical_hessian(obj_fe_only, xhat_u, eps=eps, eps_scale=eps_scale)
    H = 0.5 * (H + H.T) + 1e-12 * np.eye(idx_fe.size)

    Sigma_u = laplace_cov_from_hessian(H, eig_floor=1e-10)
    var_u = np.maximum(np.diag(Sigma_u), 0.0)

    deriv_full = d_transform_constrained_du(uparams_fe, bounds)
    deriv_fe = deriv_full[idx_fe]
    se_theta = np.abs(deriv_fe) * np.sqrt(var_u)

    return se_theta



def fe_only_full_cov_per_subject(
    s, fun, uparams_init_final, uparams_fe_final, idx_fe, idx_free, bounds,
    eps=1e-4, eps_scale=True
):
    """
    FE-only case, assuming idx_fe and idx_free are disjoint and cover ALL params.

    Computes joint Laplace covariance for packed vector:
        z = [u_fe[idx_fe], u_s[idx_free]]
    using global objective sum_t fun(theta_t, t) with:
      - FE varying (via z[:d_fe])
      - subject s free params varying (via z[d_fe:])
      - other subjects held fixed at uparams_init_final

    Returns:
      Sigma_theta_full_s : (K,K) theta-space covariance for subject s,
      ordered in the original global parameter index order (0..K-1).
    """
    idx_fe = np.asarray(idx_fe, dtype=int)
    idx_free = np.asarray(idx_free, dtype=int)

    K = uparams_init_final.shape[1]
    d_fe = idx_fe.size
    d_free = idx_free.size

    # packed-to-full index order
    idx_z = np.concatenate([idx_fe, idx_free])
    # sanity: this function assumes coverage of all dims
    # (comment out if you don't want runtime checks)
    assert idx_z.size == K
    assert np.unique(idx_z).size == K

    # base copies used in objective
    U0 = uparams_init_final.copy()   # (N,K)
    ufe0 = uparams_fe_final.copy()   # (K,)

    def obj_z(z):
        # update FE in u-space
        ufe = ufe0.copy()
        ufe[idx_fe] = z[:d_fe]

        # update subject s free params in u-space
        Us = U0.copy()
        Us[s, idx_free] = z[d_fe:]

        # FE in theta-space (full K, but only idx_fe used)
        theta_fe_full = transform_to_constrained_space(ufe, bounds)

        total = 0.0
        for t in range(Us.shape[0]):
            theta_t = transform_to_constrained_space(Us[t], bounds)
            theta_t[idx_fe] = theta_fe_full[idx_fe]
            total += fun(theta_t, t)
        return total

    # starting point at converged values
    z0 = np.concatenate([uparams_fe_final[idx_fe], uparams_init_final[s, idx_free]])

    # Hessian in u-space
    H = numerical_hessian(obj_z, z0, eps=eps, eps_scale=eps_scale)
    H = 0.5 * (H + H.T) + 1e-12 * np.eye(d_fe + d_free)

    Sigma_u = laplace_cov_from_hessian(H, eig_floor=1e-10)
    Sigma_u = 0.5 * (Sigma_u + Sigma_u.T)

    # Jacobian diag for theta transform, aligned with z
    d_fe_u = d_transform_constrained_du(uparams_fe_final, bounds)[idx_fe]
    d_free_u = d_transform_constrained_du(uparams_init_final[s], bounds)[idx_free]
    d_z = np.concatenate([d_fe_u, d_free_u])  # (K,)

    # Delta method in packed order
    Sigma_theta_z = (d_z[:, None] * Sigma_u) * d_z[None, :]
    Sigma_theta_z = 0.5 * (Sigma_theta_z + Sigma_theta_z.T)

    # Permute packed covariance into global K-order
    Sigma_theta_full = np.empty((K, K), dtype=float)
    Sigma_theta_full[np.ix_(idx_z, idx_z)] = Sigma_theta_z

    return Sigma_theta_full



def group_estimation(fun, nsubjects, params_init, bounds, idx_fe, idx_re,
                     num_cores=1, max_iter=30,
                     sigma_floor=1e-3, damping_mu=0.5, damping_sig=0.5, verbosity=1):

    params_init = np.array(params_init)
    nparams = params_init.shape[1]

    # normalize indices once
    idx_fe = np.asarray(idx_fe, dtype=int) if len(idx_fe) > 0 else np.asarray([], dtype=int)
    idx_re = np.asarray(idx_re, dtype=int) if len(idx_re) > 0 else np.asarray([], dtype=int)

    if verbosity:
        print(f'\n{SP2}Group-level optimization (MLE / MAP)')
    t0 = timeit.default_timer()

    uparams_init = transform_to_unconstrained_space(params_init, bounds)

    # ----- FE init -----
    if len(idx_fe) > 0:
        params_fe_init = np.median(params_init, axis=0)
        uparams_fe = transform_to_unconstrained_space(params_fe_init, bounds)
    else:
        uparams_fe = None

    # ----- subject-level identifiable dims -----
    all_idx = np.arange(nparams, dtype=int)
    mask_free = np.ones(nparams, dtype=bool)
    if len(idx_fe) > 0:
        mask_free[idx_fe] = False
    idx_free = all_idx[mask_free]  # subject-level identifiable dims (optimized per subject)
    d_free = idx_free.size

    # ----- RE hyperparams init -----
    has_re = (len(idx_re) > 0)
    if has_re:
        uparams_mean = uparams_init.mean(axis=0)
        uparams_sd = uparams_init.std(axis=0) + sigma_floor
    else:
        uparams_mean = None
        uparams_sd = None

    # outputs (updated each iteration; final kept)
    x_se = np.full((nsubjects, nparams), np.nan, dtype=float)
    cov_theta_free = None  # list of (d_free,d_free) from last iteration

    def subject_loop_factory(uparams_init_ref, uparams_fe_ref, uparams_mean_ref, uparams_sd_ref):

        def subject_loop(s):
            # ---- pack/unpack on free block (exclude idx_fe directions entirely) ----
            def unpack_u(u_free):
                u_full = uparams_init_ref[s].copy()
                u_full[idx_free] = u_free
                return u_full

            def obj_free(u_free):
                u_full = unpack_u(u_free)

                # Likelihood with FE imposed inside objective_empirical_bayes
                val = objective_empirical_bayes(
                    u_full, s, fun, bounds,
                    idx_re, uparams_mean_ref, uparams_sd_ref,
                    idx_fe, uparams_fe_ref
                )

                # IMPORTANT: objective_empirical_bayes already includes the RE penalty when len(idx_re)>0.
                # In FE-only, idx_re is empty, so no penalty is added.
                return val

            # ---- 1) optimize free dims ----
            x0 = uparams_init_ref[s, idx_free].copy()
            res = sciopt.minimize(obj_free, x0=x0, method='L-BFGS-B')
            u_hat_free = res.x
            u_hat_full = unpack_u(u_hat_free)

            # ---- 2) Hessian / covariance in u-space on free block ----
            H = numerical_hessian(obj_free, u_hat_free, eps=1e-4, eps_scale=True)
            H = 0.5 * (H + H.T) + 1e-12 * np.eye(d_free)

            Sigma_u_free = laplace_cov_from_hessian(H, eig_floor=1e-10)
            Sigma_u_free = 0.5 * (Sigma_u_free + Sigma_u_free.T)

            # ---- 3) delta-method covariance in theta-space on free block ----
            deriv_full = d_transform_constrained_du(u_hat_full, bounds)  # (K,)
            d = deriv_full[idx_free]                                     # (d_free,)
            Sigma_theta_free = (d[:, None] * Sigma_u_free) * d[None, :]
            Sigma_theta_free = 0.5 * (Sigma_theta_free + Sigma_theta_free.T)

            # ---- 4) SEs from theta-cov diagonal ----
            var_theta_free = np.maximum(np.diag(Sigma_theta_free), 0.0)
            se_theta_free = np.sqrt(var_theta_free)  # (d_free,)

            # ---- 5) posterior var for RE dims (only needed if has_re) ----
            if has_re:
                # idx_re is subset of idx_free because idx_fe and idx_re should be disjoint
                pos_re_in_free = np.searchsorted(idx_free, idx_re)
                var_post_re = np.maximum(np.diag(Sigma_u_free)[pos_re_in_free], 0.0)  # (d_re,)
            else:
                var_post_re = None

            return u_hat_full, var_post_re, se_theta_free, Sigma_theta_free

        return subject_loop

    # -------- main loop --------
    for it in range(max_iter):

        # ---- (A) update FE given current subject parameters ----
        if len(idx_fe) > 0:
            def objective_fe_packed(x_packed):
                uparams_tmp = uparams_fe.copy()
                uparams_tmp[idx_fe] = x_packed
                return objective_group_fe(uparams_tmp, fun, uparams_init, bounds, idx_fe)

            x0 = uparams_fe[idx_fe].copy()
            uparams_fe[idx_fe] = sciopt.minimize(objective_fe_packed, x0=x0, method='L-BFGS-B').x

        # ---- (B) update subject parameters (ALWAYS) given current FE and (if present) RE hyperparams ----
        subject_loop = subject_loop_factory(
            uparams_init_ref=uparams_init,
            uparams_fe_ref=uparams_fe,
            uparams_mean_ref=uparams_mean,
            uparams_sd_ref=uparams_sd
        )

        if num_cores > 1:
            with DillPool(num_cores) as pool:
                out = list(pool.map(subject_loop, range(nsubjects)))
        else:
            out = [subject_loop(s) for s in range(nsubjects)]

        # unpack
        uparams_init = np.array([o[0] for o in out])            # (N,K)
        se_theta_free = np.array([o[2] for o in out])           # (N,d_free)
        cov_theta_free = [o[3] for o in out]                    # list of (d_free,d_free)

        # fill SEs for free dims (includes RE and "neither")
        x_se[:, idx_free] = se_theta_free

        # ---- (C) update RE hyperparams if present ----
        if has_re:
            var_post_re = np.array([o[1] for o in out])         # (N,d_re)

            uparams_mean_new = uparams_mean.copy()
            uparams_sd_new = uparams_sd.copy()

            # mu update
            uparams_mean_new[idx_re] = uparams_init[:, idx_re].mean(axis=0)

            # sigma^2 update with curvature correction
            centered2 = (uparams_init[:, idx_re] - uparams_mean_new[idx_re]) ** 2
            sigma2_new = (centered2 + var_post_re).mean(axis=0)

            uparams_sd_new[idx_re] = np.sqrt(np.maximum(sigma2_new, sigma_floor**2))
            uparams_sd_new[idx_re] = np.maximum(uparams_sd_new[idx_re], sigma_floor)

            # damping
            uparams_mean = (1 - damping_mu) * uparams_mean + damping_mu * uparams_mean_new
            uparams_sd   = (1 - damping_sig) * uparams_sd   + damping_sig * uparams_sd_new

        if verbosity and (np.mod(it, 10) == 0):
            if has_re:
                conv = np.max(np.abs(uparams_mean - uparams_mean_new))
                convergence_str = f' (Convergence: {conv:.8f})'
            else:
                convergence_str = ''
            print(f"{TAB}{TAB}[{datetime.now().strftime('%H:%M:%S')}] Iteration {it+1} / {max_iter}{convergence_str}")

    # -------- finalize params in constrained space --------
    params = transform_to_constrained_space(uparams_init, bounds)

    # impose FE at the end (subject-level uparams_init contains arbitrary values at idx_fe)
    if len(idx_fe) > 0:
        params[:, idx_fe] = transform_to_constrained_space(uparams_fe[idx_fe], bounds[idx_fe])

        # FE SEs (global) and broadcast across subjects
        fe_se = compute_fixed_effect_se(
            fun=fun, uparams_init=uparams_init, uparams_fe=uparams_fe, bounds=bounds, idx_fe=idx_fe
        )
        x_se[:, idx_fe] = fe_se

    # -------- build result --------
    result = sciopt.OptimizeResult(x=params)
    result.x_se = x_se

    # # always provide x_cov (for free block) even in FE-only
    # result.x_cov = np.full((nsubjects, nparams, nparams), np.nan, dtype=float)
    # if cov_theta_free is not None:
    #     cov_arr = np.stack(cov_theta_free, axis=0)  # (N,d_free,d_free)
    #     # optional: enforce symmetry
    #     cov_arr = 0.5 * (cov_arr + np.swapaxes(cov_arr, 1, 2))
    #     result.x_cov[:, *np.ix_(idx_free, idx_free)] = cov_arr


    result.x_cov = np.full((nsubjects, nparams, nparams), np.nan, dtype=float)
    if has_re:
        # keep your existing behavior in RE case (free-free only)
        cov_arr = np.stack(cov_theta_free, axis=0)  # (N,d_free,d_free)
        cov_arr = 0.5 * (cov_arr + np.swapaxes(cov_arr, 1, 2))
        # result.x_cov[:, *np.ix_(idx_free, idx_free)] = cov_arr  # does not work in Python 3.10
        result.x_cov[(slice(None),) + np.ix_(idx_free, idx_free)] = cov_arr
    else:
        # FE-only: compute full KxK per subject (includes FE blocks + cross)
        if num_cores > 1:
            with DillPool(num_cores) as pool:
                cov_full_list = list(pool.map(
                    lambda s: fe_only_full_cov_per_subject(s, fun, uparams_init, uparams_fe, idx_fe, idx_free, bounds),
                    range(nsubjects)
                ))
        else:
            cov_full_list = [fe_only_full_cov_per_subject(s, fun, uparams_init, uparams_fe, idx_fe, idx_free, bounds)
                             for s in range(nsubjects)]

        result.x_cov = np.stack(cov_full_list, axis=0)  # (N,K,K)

    # RE population summary only if RE exists
    if has_re:
        result.x_re_pop_mean_sd = population_summary(uparams_mean, uparams_sd, bounds, idx_re)
    else:
        result.x_re_pop_mean_sd = None

    result.execution_time = timeit.default_timer() - t0
    if verbosity:
        print(f'{TAB}.. finished ({result.execution_time:.1f} secs).')

    return result


def numerical_hessian(phi, x, eps=1e-4, eps_scale=False):
    """
    Central-difference Hessian.

    eps:
      - float: fixed step size for all dims
      - array-like (d,): per-dim step sizes
    eps_scale:
      - if True and eps is a float, use eps_i = eps*(|x_i|+1)

    phi: callable R^d -> R
    x: (d,)
    """
    x = np.asarray(x, float)
    d = x.size
    H = np.zeros((d, d), float)
    fx = phi(x)

    if np.isscalar(eps):
        if eps_scale:
            eps_vec = eps * (np.abs(x) + 1.0)
        else:
            eps_vec = np.full(d, float(eps))
    else:
        eps_vec = np.asarray(eps, float)
        if eps_vec.shape != (d,):
            raise ValueError("eps must be scalar or shape (d,)")

    for i in range(d):
        ei = np.zeros(d); ei[i] = eps_vec[i]
        f_ip = phi(x + ei)
        f_im = phi(x - ei)
        H[i, i] = (f_ip - 2 * fx + f_im) / (eps_vec[i] ** 2)

        for j in range(i + 1, d):
            ej = np.zeros(d); ej[j] = eps_vec[j]
            f_pp = phi(x + ei + ej)
            f_pm = phi(x + ei - ej)
            f_mp = phi(x - ei + ej)
            f_mm = phi(x - ei - ej)
            H_ij = (f_pp - f_pm - f_mp + f_mm) / (4 * eps_vec[i] * eps_vec[j])
            H[i, j] = H_ij
            H[j, i] = H_ij

    return H


def objective_empirical_bayes(uparams, sub_ind, fun, bounds, idx_re, uparams_re_mean, uparams_re_sd, idx_fe, uparams_fe):
    params = transform_to_constrained_space(uparams, bounds)
    if len(idx_fe) > 0:
        params_fe = transform_to_constrained_space(uparams_fe[idx_fe], bounds[idx_fe])
        params[idx_fe] = params_fe
    val = fun(params, sub_ind)
    # Gaussian prior penalty on random-effects
    if len(idx_re) > 0:
        val += 0.5 * np.sum((uparams[idx_re] - uparams_re_mean[idx_re])**2 / uparams_re_sd[idx_re]**2)
    return val


def objective_group_fe(uparams_fe, fun, uparams, bounds, idx_fe):
    # (uparams_tmp, fun, uparams_re_init, bounds, idx_fe)
    negll = 0.0
    params_fe = transform_to_constrained_space(uparams_fe, bounds)  # (K,)
    for s in range(uparams.shape[0]):
        params = transform_to_constrained_space(uparams[s], bounds)
        params[idx_fe] = params_fe[idx_fe]
        negll += fun(params, s)
    return negll


def population_summary(uparams_re_mean, uparams_re_sd, bounds, idx_re, n_draws=200000, seed=0):


    # params_re_popmean, params_re_popsd = population_summary(uparams_re_mean, uparams_re_sd, bounds, idx_re)

    """
    Compute group-level summaries in params-space using Monte Carlo sampling based on:
        z_j ~ Normal(mu_j, sigma_j)  (for j in re_idx)
        params = transform_constrained(uparams, bounds)

    Returns summaries only for re_idx parameters.
    """
    rng = np.random.default_rng(seed)

    # Draw uparams for RE dims only: (n_draws, |re_idx|)
    uparams_re = rng.normal(loc=uparams_re_mean[idx_re], scale=uparams_re_sd[idx_re], size=(n_draws, idx_re.size))

    # Put into full K-dim uparams with fixed values for non-RE dims (mu is a reasonable plug-in)
    uparams = np.tile(uparams_re_mean, (n_draws, 1))
    uparams[:, idx_re] = uparams_re

    # Transform to params-space
    params = transform_to_constrained_space(uparams, bounds)  # expects (n_draws, K) support

    # Summarize only RE dims
    params_re = params[:, idx_re]

    theta_mean, theta_sd = params_re.mean(axis=0), params_re.std(axis=0, ddof=0)

    return theta_mean, theta_sd


def transform_to_constrained_space(uparams, bounds):
    lower, upper = bounds.T
    uparams = np.asarray(uparams, dtype=float)

    params = np.empty_like(uparams, dtype=float)

    m_unbounded  = np.isinf(lower) & np.isinf(upper)
    m_lower_only = np.isfinite(lower) & np.isinf(upper)
    m_upper_only = np.isinf(lower) & np.isfinite(upper)
    m_bounded    = np.isfinite(lower) & np.isfinite(upper)

    # Use ... to apply along last axis
    params[..., m_unbounded]  = uparams[..., m_unbounded]

    # Optional: clip to avoid exp overflow in the one-sided cases too
    # (exp(709) ~ 8e307 is near float64 max)
    z = np.clip(uparams, -709.0, 709.0)

    params[..., m_lower_only] = lower[m_lower_only] + np.exp(z[..., m_lower_only])
    params[..., m_upper_only] = upper[m_upper_only] - np.exp(z[..., m_upper_only])

    # Stable bounded transform
    params[..., m_bounded] = lower[m_bounded] + (upper[m_bounded] - lower[m_bounded]) * expit(uparams[..., m_bounded])
    return params


def transform_to_unconstrained_space(params, bounds, eps=1e-12):
    lower, upper = bounds.T
    params = np.asarray(params, dtype=float)

    uparams = np.empty_like(params, dtype=float)

    m_unbounded  = np.isinf(lower) & np.isinf(upper)
    m_lower_only = np.isfinite(lower) & np.isinf(upper)
    m_upper_only = np.isinf(lower) & np.isfinite(upper)
    m_bounded    = np.isfinite(lower) & np.isfinite(upper)

    uparams[..., m_unbounded] = params[..., m_unbounded]

    # Guard against exact-bound values
    uparams[..., m_lower_only] = np.log(np.maximum(params[..., m_lower_only] - lower[m_lower_only], eps))
    uparams[..., m_upper_only] = np.log(np.maximum(upper[m_upper_only] - params[..., m_upper_only], eps))

    x = (params[..., m_bounded] - lower[m_bounded]) / (upper[m_bounded] - lower[m_bounded])
    x = np.clip(x, eps, 1 - eps)
    uparams[..., m_bounded] = np.log(x) - np.log1p(-x)

    return uparams


class RandomDisplacementBoundsConstraints(object):
    def __init__(self, bounds, constraints, stepsize=0.5):
        self.xmin = bounds.lb
        self.xmax = bounds.ub
        self.constraints = constraints
        self.stepsize = stepsize

    def __call__(self, x):
        while True:
            xnew = x + np.random.uniform(-self.stepsize, self.stepsize, np.shape(x))
            if np.all(xnew < self.xmax) and np.all(xnew > self.xmin) and (self.constraints is None or (np.all(
                    [con['fun'](xnew) >= 0 for con in self.constraints]))):
                break
        return xnew


class BoundsConstraints(object):

    def __init__(self, bounds, constraints=None):
        self.xmin = bounds.lb
        self.xmax = bounds.ub
        self.constraints = constraints

    def __call__(self, **kwargs):
        x = kwargs["x_new"]
        tmax = bool(np.all(x <= self.xmax))
        tmin = bool(np.all(x >= self.xmin))
        cons = self.constraints is None or bool(np.all([con['fun'](x) >= 0 for con in self.constraints]))
        return tmax and tmin and cons



from .model import ReMeta, load_dataset
from .configuration import Configuration
from .simulation import simulate
from .plotting import plot_psychometric, plot_confidence_histogram, plot_stimulus_versus_confidence



from __future__ import annotations

import matplotlib.pyplot as plt
import matplotlib.text as mtext
import numpy as np
import warnings
from matplotlib.ticker import FormatStrFormatter
from scipy.special import ndtr
from scipy.stats import sem

from remeta.simulation import simulate, Simulation
from remeta.model import Configuration
from remeta.util import _check_param, listlike, fmp

color_generative_type2 = np.array([231, 168, 116]) / 255
color_generative_type2b = np.array([47, 158, 47]) / 255
color_data = [0.6, 0.6, 0.6]

color_inc = [0.87, 0.27, 0.33]
color_cor = [0, 0.67, 0.5]

color_model = (0.55, 0.55, 0.69)
color_model_wrong = np.array([152, 75, 75]) / 255

symbols = dict(
    type1_noise=r'$\sigma_1$',
    type1_noise_heteroscedastic=r'$\sigma_\mathrm{1,1}$',
    type1_thresh=r'$\vartheta_1$',
    type1_bias=r'$\delta_1$',
    type2_noise=r'$\sigma_2$',
    type2_evidence_bias=r'$\varphi_2$',
    type2_confidence_bias=r'$\alpha_2$',
    type2_criteria=r'$\gamma^\mathrm{i}$'
)

FONTSIZE = dict(
    label=14,
    xlabel=14,
    ylabel=14,
    tick=11,
    xtick=11,
    ytick=11,
    title=14
)


class _LegendTitle(object):
    def __init__(self, text_props=None):
        self.text_props = text_props or {}
        super(_LegendTitle, self).__init__()

    def legend_artist(self, legend, orig_handle, fontsize, handlebox):  # noqa
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        title = mtext.Text(x0, y0, orig_handle, **self.text_props)
        # title = mtext.Text(x0, y0, r'\underline{' + orig_handle + '}', usetex=True, **self.text_props)
        handlebox.add_artist(title)
        return title


def _logistic(x, sigma, thresh, bias):
    beta = np.pi / (np.sqrt(3) * sigma)
    return \
        (np.abs(x) >= thresh) * (
                1 / (1 + np.exp(-beta * (x + bias)))) + \
        (np.abs(x) < thresh) * (1 / (1 + np.exp(-beta * bias)))


def _normal(x, sigma, thresh, bias):
    return \
        (np.abs(x) >= thresh) * ndtr((x + bias) / sigma) + \
        (np.abs(x) < thresh) * ndtr(bias / sigma)


def _linear(x, thresh, bias):
    y = (np.abs(x) > thresh) * (x - np.sign(x) * thresh) + bias
    return y



def _tanh(x, beta, thresh, offset):
    return \
        (np.abs(x) > thresh) * (
                (1 - offset) * np.tanh(beta * (x - np.sign(x) * thresh)) + np.sign(x) * offset) + \
        (np.abs(x) <= thresh) * np.sign(x) * offset


def plot_psychometric(
        stimuli_or_simulation: list[float] | np.typing.NDArray[float] | Simulation | None = None,
        choices: list[float] | np.typing.NDArray[float] | None = None,
        type1_noise: float | list[float] | None = None,
        type1_bias: float | None = None,
        type1_thresh: float | list[float] | None = None,
        stim_max: float | None = None,
        params: dict | None = None,
        type1_noise_dist: str = 'normal',
        cfg: Configuration | None = None,
        model_prediction: bool = False,
        model_only: bool = False,
        highlight_fit: bool = False,
        errorbars: bool = True,
        path_export: str | None = None
) -> None:
    """Plot psychometric function.

    Usage:
        **Psychometric curve for empirical data:**

        `plot_psychometric(stimuli, choices)`  # Data only

        `plot_psychometric(stimuli, choices, type1_noise, type1_bias, ...)`  # Data + model

        `plot_psychometric(stimuli, choices, params=params)`  # Data + model

        **Psychometric curve for a simulation instance: (`remeta.simulation.Simulation`):**

        `plot_psychometric(simulation)`  # (Simulated) Data

        `plot_psychometric(simulation, model_prediction=True)`  # (Simulated) Data + model

        **Model only:**

        `plot_psychometric(type1_noise=type1_noise, type1_bias=type1_bias, ...)`  # Model

        `plot_psychometric(params)`  # Model

    Args:
        stimuli_or_simulation: 1d stimulus array (normalized to [-1; 1]) or [Simulation][remeta.simulation.Simulation] object
        choices: 1d choice array
        type1_noise: Type 1 noise.
        type1_bias: Type 1 bias.
        type1_thresh: Type 1 threshold.
        stim_max: maximum stimulus intensity.
        params: instead of passing parameters as separate parameters, one can pass a parameter dictionary
        cfg: If a [Configuration][remeta.configuration.Configuration] instance is passed, checks are performed for
             `cfg.param_type1_bias.enable` and `cfg.param_type1_thresh.enable`
         model_prediction: whether to show model-predicted values for comparison
        model_only: Show the model prediction only (auto-set to True if no data are passed)
        highlight_fit: Emphasize the fit over the data
        errorbars: Whether to include error bars for empirical data (SD)
    """

    if type1_noise is not None or params is not None:
        model_prediction = True

    if stimuli_or_simulation is not None and choices is None:  # assume a dataset is passed
        cfg = stimuli_or_simulation.cfg
        stimuli = stimuli_or_simulation.stimuli
        choices = stimuli_or_simulation.choices
        if model_prediction:
            params = stimuli_or_simulation.params
            type1_noise = params['type1_noise']
            type1_bias = params['type1_bias'] if cfg.param_type1_bias.enable else None
            type1_thresh = params['type1_thresh'] if cfg.param_type1_thresh.enable else None
    else:
        if choices is None:
            model_only = True
        stimuli = None if model_only else stimuli_or_simulation
        if model_prediction:
            if params is None:
                if type1_noise is None:
                    raise ValueError('Model predictions requested, but type 1 noise is unspecified.')
                params = dict(type1_noise=type1_noise)
            else:
                type1_noise = params['type1_noise']
                if 'type1_bias' in params and not (cfg is not None and not cfg.param_type1_bias.enable):
                    type1_bias = params['type1_bias']
                if 'type1_thresh' in params and not (cfg is not None and not cfg.param_type1_thresh.enable):
                    type1_thresh = params['type1_thresh']

    if stim_max is None:
        stim_max = 1 if stimuli is None else np.max(np.abs(stimuli))

    xrange_neg = np.linspace(-stim_max, 0, 1000)
    xrange_pos = np.linspace(stim_max / 1000, stim_max, 1000)

    if model_prediction:
        type1_noise = _check_param(type1_noise)
        if (cfg is None and type1_thresh is not None) or (cfg is not None and cfg.param_type1_thresh.enable):
            params['type1_thresh'] = type1_thresh
            type1_thresh = _check_param(type1_thresh)
        else:
            type1_thresh = [0, 0]

        if (cfg is None and type1_bias is not None) or (cfg is not None and cfg.param_type1_bias.enable):
            params['type1_bias'] = type1_bias
            type1_bias = _check_param(type1_bias)
        else:
            type1_bias = [0, 0]

        if cfg is not None:
            type1_noise_dist = cfg.param_type1_noise.model
        cdf = _logistic if type1_noise_dist == 'logistic' else _normal
        posterior_neg = cdf(xrange_neg, type1_noise[0], type1_thresh[0], type1_bias[0])
        posterior_pos = cdf(xrange_pos, type1_noise[1], type1_thresh[1], type1_bias[1])


    fig = plt.figure(figsize=(6, 3.5))
    fig.subplots_adjust(bottom=0.2)
    gs = fig.add_gridspec(1, 2, width_ratios=[1, 0.5], wspace=0)
    ax = fig.add_subplot(gs[0, 0])
    ax_leg = fig.add_subplot(gs[0, 1])
    ax_leg.axis("off")
    plt.sca(ax)


    if not model_only:
        stimulus_ids = (stimuli > 0).astype(int)
        levels = np.unique(stimuli)
        choiceprob_neg = np.array([np.mean(choices[(stimuli == v) & (stimulus_ids == 0)] ==
                                           stimulus_ids[(stimuli == v) & (stimulus_ids == 0)])
                                   for v in levels[levels < 0]])
        choiceprob_pos = np.array([np.mean(choices[(stimuli == v) & (stimulus_ids == 1)] ==
                                           stimulus_ids[(stimuli == v) & (stimulus_ids == 1)])
                                   for v in levels[levels > 0]])
        if errorbars:
            choiceprob_neg_se = np.array([sem(choices[(stimuli == v) & (stimulus_ids == 0)] ==
                                               stimulus_ids[(stimuli == v) & (stimulus_ids == 0)])
                                       for v in levels[levels < 0]])
            choiceprob_pos_se = np.array([sem(choices[(stimuli == v) & (stimulus_ids == 1)] ==
                                               stimulus_ids[(stimuli == v) & (stimulus_ids == 1)])
                                       for v in levels[levels > 0]])
        else:
            choiceprob_neg_se, choiceprob_pos_se = np.nan, np.nan
        plt.errorbar(levels[levels < 0], 1-choiceprob_neg, yerr=choiceprob_neg_se, fmt='o', color=color_data, mec='k', ls='' if model_prediction else '-',
                     label=f'Data $S^-$ (mean{["", "±SE"][int(errorbars)]})', clip_on=False, zorder=11, alpha=(1, 0.2)[highlight_fit])
        if not model_prediction:
            plt.plot([levels[levels < 0][-1], levels[levels > 0][0]], [1-choiceprob_neg[-1], choiceprob_pos[0]], color=color_data)
        plt.errorbar(levels[levels > 0], choiceprob_pos, yerr=choiceprob_pos_se, fmt='s', color=color_data, mec='k', ls='' if model_prediction else '-',
                     label=f'Data $S^+$ (mean{["", "±SE"][int(errorbars)]})', clip_on=False, zorder=11, alpha=(1, 0.2)[highlight_fit])

    if model_prediction:
        plt.plot(xrange_neg, posterior_neg, '-', lw=(2, 5)[highlight_fit], color=color_model, clip_on=False,
                 zorder=(10, 12)[highlight_fit], label=f'Model')
        plt.plot(xrange_pos, posterior_pos, '-', lw=(2, 5)[highlight_fit], color=color_model, clip_on=False,
                 zorder=(10, 12)[highlight_fit])

    plt.plot([-stim_max, stim_max], [0.5, 0.5], 'k-', lw=0.5)
    plt.plot([0, 0], [-0.02, 1.02], 'k-', lw=0.5)

    ax.yaxis.grid('on', color=[0.9, 0.9, 0.9])
    plt.xlim((-stim_max, stim_max))
    plt.ylim((0, 1))
    plt.xlabel('Stimulus ($x$)')
    plt.ylabel('Choice probability $S^+$')
    ax.xaxis.set_major_formatter(FormatStrFormatter('%.2g'))

    if model_prediction:
        anot_type1 = []
        for i, (k, v) in enumerate(params.items()):
            # if (cfg is None and k in params) or (cfg is not None and getattr(cfg, f"enable_param_{k}")) and k.startswith('type1_'):
            if k.startswith('type1') and (cfg is None or (cfg is not None and getattr(cfg, f"param_{k}").enable)):
                if hasattr(v, '__len__'):
                    val = ', '.join([fmp(p) for p in v])
                    anot_type1 += [f"${symbols[k][1:-1]}=" + f"[{val}]$"]
                else:
                    anot_type1 += [f"${symbols[k][1:-1]}={fmp(v)}$"]
        plt.text(1.045, 0.1, r'Parameters:' + '\n' + '\n'.join(anot_type1), transform=ax.transAxes,
                 bbox=dict(fc=[1, 1, 1], ec=[0.5, 0.5, 0.5], lw=1, pad=5), fontsize=11)
    set_fontsize(label='default', tick='default')

    # ax_leg.legend(bbox_to_anchor=(1.01, 1), loc="upper left", fontsize=11, handlelength=1)
    ax_leg.legend(*ax.get_legend_handles_labels(), loc="upper left", fontsize=11, handlelength=1)

    if path_export is not None:
        plt.savefig(path_export, bbox_inches='tight', pad_inches=0.02)


def plot_stimulus_versus_confidence(
        stimuli_or_simulation: list[float] | np.typing.NDArray[float] | Simulation | None = None,
        confidence: list[float] | np.typing.NDArray[float] | None = None,
        choices: list[float] | np.typing.NDArray[float] | None = None,
        params: dict | None = None,
        type1_noise: float | list[float] | None = None,
        type1_bias: float | list[float] | None = None,
        type1_thresh: float | list[float] | None = None,
        type2_noise: float | None = None,
        type2_evidence_bias_mult: float | None = None,
        type2_criteria: float | list[float] | None = None,
        cfg: Configuration | None = None,
        stim_max: float | None = None,
        errorbar_type: str | None = 'SEM',
        probability_correct: bool = False,
        separate_by_accuracy: bool = False,
        model_prediction: bool = False,
        model_prediction_nsamples: int = 10000,
        model_only: bool = False,
        model_prediction_disable_type2_noise: bool = False,
        path_export: str | None = None
) -> None:
    """ Plot the relationship between stimulus levels and confidence.

    Usage:
        **Plot for empirical data:**

        `plot_stimulus_versus_confidence(stimuli, choices)`  # Data

        `plot_stimulus_versus_confidence(stimuli, choices, type1_noise, type2_noise, ...)`  # Data + Model

        `plot_stimulus_versus_confidence(stimuli, choices, params, ...)`  # Data + Model

        **Plot for simulation instance: (`remeta.simulation.Simulation`):**

        `plot_stimulus_versus_confidence(simulation)`  # (Simulated) Data

        `plot_stimulus_versus_confidence(simulation, model_prediction=True)`  # (Simulated) Data + model

        **Model only:**

        `plot_confidence_histogram(type1_noise=type1_noise, type2_noise=type2_noise, ...)`  # Model

        `plot_stimulus_versus_confidence(params)`  # Model

    Args:
        stimuli_or_simulation: 1d stimulus array (normalized to [-1; 1]) or [Simulation][remeta.simulation.Simulation] object
        confidence: 1d confidence array
        choices: 1d choice array
        params: pass parameters as a dictionary
        type1_noise: Type 1 noise.
        type1_bias: Type 1 bias.
        type1_thresh: Type 1 threshold.
        type2_noise: Metacognitive noise. Required
        type2_evidence_bias_mult: Multiplicative metacognitive bias
        type2_criteria: Confidence criteria
        cfg: Place holder - checking the configuration object is not yet implemented
        stim_max: maximum stimulus intensity
        errorbar_type: either None (disable), 'SD' (standard deviation) or 'SEM' (standard error)
        probability_correct: if True, convert confidence (range 0-1) to subjective probability correct (range 0.5-1),
        separate_by_accuracy: separate plots for correct and incorrect responses
        model_prediction: whether to show model-predicted values for comparison
        model_prediction_nsamples: number of samples used to generate model predictions
        model_only: Show the model prediction only (auto-set to True if no data are passed)
        model_prediction_disable_type2_noise: plot model prediction with ~0 metacognitive noise
    """

    if stimuli_or_simulation is not None and confidence is None:  # assume first parameter is a simulated dataset
        from copy import deepcopy
        cfg = deepcopy(stimuli_or_simulation.cfg)
        stimuli, confidence, choices = stimuli_or_simulation.stimuli, stimuli_or_simulation.confidence, stimuli_or_simulation.choices
        params = stimuli_or_simulation.params.copy()
    else:

        if separate_by_accuracy and choices is None and not model_only:
            raise ValueError('If separate_by_accuracy is True, choices must be passed.')

        if confidence is None or params is not None or (type1_noise is not None and type2_noise is not None):
            model_prediction = True
        stimuli = None if model_only else stimuli_or_simulation
        choices = None if model_only else choices
        if stimuli is None:
            model_only = True
        if model_prediction:
            if params is None:
                if type1_noise is None:
                    raise ValueError('Type 1 noise is unspecified.')
                if type2_noise is None:
                    raise ValueError('Type 2 noise is unspecified.')
                params = dict(type1_noise=type1_noise, type2_noise=type2_noise)
                for param in ('type1_bias', 'type1_thresh', 'type2_evidence_bias_mult', 'type2_criteria'):
                    if (value := eval(param)) is not None and not (cfg is not None and not getattr(cfg, f'param_{param}').enable):
                        params[param] = value
            else:
                params = params.copy()
                if 'type1_noise' not in params:
                    raise ValueError('Type 1 noise is unspecified.')
                if 'type2_noise' not in params:
                    raise ValueError('Type 2 noise is unspecified.')
                for param in ('type1_bias', 'type1_thresh', 'type2_evidence_bias_mult', 'type2_criteria'):
                    if param in params and (cfg is not None and not getattr(cfg, f'param_{param}').enable):
                        params.pop(param)

    if model_prediction and model_prediction_disable_type2_noise:
        params['type2_noise'] = 1e-5


    if probability_correct:
        confidence = (confidence + 1) / 2

    fig = plt.figure(figsize=(6, 3.5))
    fig.subplots_adjust(bottom=0.2)
    gs = fig.add_gridspec(1, 2, width_ratios=[1, 0.5], wspace=0)
    ax = fig.add_subplot(gs[0, 0])
    ax_leg = fig.add_subplot(gs[0, 1])
    ax_leg.axis("off")
    plt.sca(ax)
    ax.xaxis.set_major_formatter(FormatStrFormatter('%.2g'))

    if stim_max is None:
        stim_max = 1 if stimuli is None else np.max(np.abs(stimuli))


    if errorbar_type is not None:
        if errorbar_type == 'SD':
            errorfun = np.std
        elif errorbar_type == 'SEM':
            errorfun = sem
        else:
            raise ValueError(f"Unknown errorbar type '{errorbar_type}' (valid: 'SD', 'SEM')")

    chance_level_ref = 0.75 if probability_correct else 0.5
    if not separate_by_accuracy:
        plt.plot([-1.05*stim_max, 1.05*stim_max], [chance_level_ref, chance_level_ref], lw=1, color=[0.2, 0.2, 0.2],
                 ls=':', label='Theoretical value\nat chance level')
    plt.plot([0, 0], [0, 1], 'k-', lw=0.5)

    if not model_only:
        stim_levels = sorted(np.unique(stimuli))
        if separate_by_accuracy:
            accuracy = np.sign(stimuli) == np.sign(choices - 0.5)
            conf_av_inc = np.array([np.nan if (cnd :=(~accuracy & (stimuli == stim_level))).sum() == 0 else np.mean(confidence[cnd]) for stim_level in stim_levels])
            conf_av_cor = np.array([np.nan if (cnd :=(accuracy & (stimuli == stim_level))).sum() == 0 else np.mean(confidence[cnd]) for stim_level in stim_levels])
            if errorbar_type is not None:
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore', category=RuntimeWarning)
                    conf_err_inc = np.array([np.nan if (cnd :=(~accuracy & (stimuli == stim_level))).sum() == 0 else errorfun(confidence[cnd]) for stim_level in stim_levels])
                    conf_err_cor = np.array([np.nan if (cnd :=(accuracy & (stimuli == stim_level))).sum() == 0 else errorfun(confidence[cnd]) for stim_level in stim_levels])
            else:
                conf_err_inc = np.zeros(len(stim_levels))
                conf_err_cor = np.zeros(len(stim_levels))
            plt.errorbar(
                stim_levels,
                conf_av_cor,
                yerr=None if errorbar_type is None else conf_err_cor,
                marker='o', markersize=5, mew=1, mec='k', color='None', ecolor=color_cor, mfc=color_cor, clip_on=False,
                elinewidth=1.5, capsize=5, label=f'Data (correct,\nmean{"" if errorbar_type is None else f"±{errorbar_type}"})'
            )
            plt.errorbar(
                stim_levels,
                conf_av_inc,
                yerr=None if errorbar_type is None else conf_err_inc,
                marker='o', markersize=5, mew=1, mec='k', color='None', ecolor=color_inc, mfc=color_inc, clip_on=False,
                elinewidth=1.5, capsize=5, label=f'Data (incorrect,\nmean{"" if errorbar_type is None else f"±{errorbar_type}"})'
            )
            conf_min_data = min(np.nanmin(conf_av_inc - conf_err_inc), np.nanmin(conf_av_cor - conf_err_cor))
        else:
            conf_av = np.array([np.mean(confidence[stimuli == stim_level]) for stim_level in stim_levels])
            if errorbar_type is not None:
                conf_err = np.array([errorfun(confidence[stimuli == stim_level]) for stim_level in stim_levels])
            else:
                conf_err = np.zeros(len(stim_levels))

            plt.errorbar(
                stim_levels,
                conf_av,
                ls='' if model_prediction else '-',
                yerr=None if errorbar_type is None else conf_err,
                marker='o', markersize=5, mew=1, mec='k', color=color_data, ecolor='k', mfc=color_data, clip_on=False,
                elinewidth=1.5, capsize=5, label=f'Data (mean{"" if errorbar_type is None else f"±{errorbar_type}"})'
            )
            conf_min_data = np.min(conf_av - conf_err)

    if model_prediction:

        # nlevels = 100
        # nrows_per_level = int(np.ceil(model_prediction_nsamples / nlevels / 2))
        # stim_min = stim_max / nlevels
        # levels = np.hstack((np.linspace(-stim_max, -stim_min, nlevels), np.linspace(stim_min, stim_max, nlevels)))
        # x_stim = np.tile(levels, (nrows_per_level, 1))
        #
        # y_decval = stimulus_to_decision_value(x_stim, params, return_only_decval=True)
        # c_conf = type1_evidence_to_confidence(
        #     z1_type1_evidence=np.abs(y_decval), y_decval=y_decval,
        #     **params
        # ).mean(axis=0)

        _nsamples = int(np.ceil(model_prediction_nsamples / 2))
        stim_min = stim_max / _nsamples
        levels = np.hstack((np.linspace(-stim_max, -stim_min, _nsamples), np.linspace(stim_min, stim_max, _nsamples)))
        # y_decval = stimulus_to_decision_value(levels, params, return_only_decval=True)
        # c_conf = type1_evidence_to_confidence(
        #     z1_type1_evidence=np.abs(y_decval), x_stim=levels,
        #     **params
        # )
        ds = simulate(
            nsubjects=100,
            params=params, cfg=cfg, custom_stimuli=levels, verbosity=False,
            stim_max=stim_max, squeeze=True, compute_stats=False,
            silence_warnings=True
        )
        c_conf = (ds.confidence + 1) / 2 if probability_correct else ds.confidence

        # from scipy.interpolate import UnivariateSpline

        if 'type1_bias' in params:
            if listlike(params['type1_bias']):
                warnings.warn(f'Stimulus-dependent bias is currently not supported for this plot.')
                bias = params['type1_bias'][0]
            else:
                bias = params['type1_bias']
        else:
            bias = 0

        # Constrain the spline in a way that it is exactly chance_level_ref for x_stim = -bias
        # We do this by weighting an additional datapoint at x_stim = -bias very highly
        # ind_levels_min = np.argmin(np.abs(levels + bias))
        # # insert data point at the position that is closest to already present values, while not
        # # preserving ordering
        # insert_idx = ind_levels_min + 1 if (levels[ind_levels_min] + bias < 0) else ind_levels_min
        # weights = np.insert(np.ones_like(c_conf), insert_idx, 1e8)
        # c_conf_aug = np.insert(c_conf, insert_idx, chance_level_ref)
        # levels_aug = np.insert(levels, insert_idx, -bias)
        # c_conf_final = UnivariateSpline(levels_aug, c_conf_aug, k=3, w=weights)(levels)

        from scipy.ndimage import gaussian_filter1d
        if separate_by_accuracy:
            accuracy = np.sign(ds.choices - 0.5) == np.sign(ds.stimuli)
            c_conf_inc_, c_conf_cor_ = c_conf.copy(), c_conf.copy()
            c_conf_inc_[accuracy] = np.nan
            c_conf_cor_[~accuracy] = np.nan
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', category=RuntimeWarning)
                c_conf_inc_mean, c_conf_cor_mean = np.nanmean(c_conf_inc_, axis=0), np.nanmean(c_conf_cor_, axis=0)
            # Interpolate nans
            isnan_inc, isnan_cor = np.isnan(c_conf_inc_mean), np.isnan(c_conf_cor_mean)
            c_conf_inc_mean[isnan_inc] = np.interp(np.arange(2*_nsamples)[isnan_inc], np.arange(2*_nsamples)[~isnan_inc], c_conf_inc_mean[~isnan_inc])
            c_conf_cor_mean[isnan_cor] = np.interp(np.arange(2*_nsamples)[isnan_cor], np.arange(2*_nsamples)[~isnan_cor], c_conf_cor_mean[~isnan_cor])
            c_conf_inc = gaussian_filter1d(c_conf_inc_mean, sigma=model_prediction_nsamples/20)
            c_conf_cor = gaussian_filter1d(c_conf_cor_mean, sigma=model_prediction_nsamples/20)
            plt.plot(levels, c_conf_cor, '-', lw=2, color=color_cor, clip_on=False,
                     label='Model prediction\n(correct)', alpha=0.5)
                     # label='Model prediction' + (r"($\sigma_2=0$)" if model_prediction_disable_type2_noise else "")+'\n(correct)', alpha=0.5)
            plt.plot(levels, c_conf_inc, '-', lw=2, color=color_inc, clip_on=False,
                     label='Model prediction\n(incorrect)', alpha=0.5)
                     # label='Model prediction' + (r"($\sigma_2=0$)" if model_prediction_disable_type2_noise else "")+'\n(incorrect)', alpha=0.5)
            conf_min_model = min(min(c_conf_inc), min(c_conf_cor))
        else:
            c_conf_final = gaussian_filter1d(c_conf.mean(axis=0), sigma=model_prediction_nsamples/20)
            plt.plot(levels, c_conf_final, '-', lw=2, color=color_model, clip_on=False,
                     label=f'Model prediction')
                     # label=f'Model prediction' + (r"($\sigma_2=0$)" if model_prediction_disable_type2_noise else ""))
            conf_min_model = min(c_conf_final)

    plt.xlim(-1.05*stim_max, 1.05*stim_max)
    conf_min = conf_min_model if model_only else (min(conf_min_data, conf_min_model) if model_prediction else conf_min_data)
    step = 0.05 if probability_correct else 0.1
    ymin = (np.floor(conf_min / step) * step) if conf_min < chance_level_ref else chance_level_ref - step / 2
    plt.ylim(ymin, 1)
    plt.xlabel('Stimulus ($x$)')
    plt.ylabel('Subjective prob. correct' if probability_correct else 'Confidence ($C$)')

    if model_prediction:
        anot_type2 = []
        for i, (k, v) in enumerate(params.items()):
            if k.startswith('type2_'):
                if listlike(v):
                    val = ', '.join([fmp(p) for p in v])
                    anot_type2 += [f"${symbols[k][1:-1]}=" + f"[{val}]$"]
                else:
                    anot_type2 += [f"${symbols[k][1:-1]}={fmp(v)}$"]
        plt.text(1.045, 0.1-0.2*separate_by_accuracy, r'Type 2 parameters:' + '\n' + '\n'.join(anot_type2), transform=ax.transAxes,
                 bbox=dict(fc=[1, 1, 1], ec=[0.5, 0.5, 0.5], lw=1, pad=5), fontsize=11)

    # plt.legend(bbox_to_anchor=(1.01, 1), loc="upper left", fontsize=11, handlelength=1)
    ax_leg.legend(*ax.get_legend_handles_labels(), loc="upper left", fontsize=11, handlelength=1)

    set_fontsize(label='default', tick='default')
    if path_export is not None:
        plt.savefig(path_export, bbox_inches='tight', pad_inches=0.02)




def plot_confidence_histogram(
        confidence_or_simulation: list[float] | np.typing.NDArray[float] | Simulation | None = None,
        stimuli: list[float] | np.typing.NDArray[float] | None = None,
        choices: list[float] | np.typing.NDArray[float] | None = None,
        params: dict[str, float] | None = None,
        type1_noise: float | list[float] | None = None,
        type1_bias: float | list[float] | None = None,
        type1_thresh: float | list[float] | None = None,
        type2_noise: float | None = None,
        type2_evidence_bias_mult: float | None = None,
        type2_criteria: list[float] | None = None,
        cfg: Configuration | None = None,
        stim_max: float | None = None,
        probability_correct: bool = False,
        model_prediction: bool = False,
        model_only: bool = False,
        separate_by_category: bool = False,
        separate_by_accuracy: bool = False,
        model_prediction_nsamples: int = 10000,
        path_export: str | None = None
) -> None:
    """ Plot the relationship between stimulus levels and confidence.

    Usage:
        **Plot for empirical data:**

        `plot_confidence_histogram(confidence)`  # Data

        `plot_confidence_histogram(confidence, stimuli, choices, separate_by_accuracy=True)`  # Data

        `plot_confidence_histogram(confidence, type1_noise, type2_noise, ...)`  # Data + Model

        `plot_confidence_histogram(confidence, params, ...)`  # Data + Model

        **Plot for simulation instance: (`remeta.simulation.Simulation`):**

        `plot_confidence_histogram(simulation)`  # (Simulated) Data

        `plot_confidence_histogram(simulation, model_prediction=True)`  # (Simulated) Data + model

        **Model only:**

        `plot_confidence_histogram(type1_noise=type1_noise, type2_noise=type2_noise, ...)`  # Model

        `plot_confidence_histogram(params)`  # Model

    Args:
        confidence_or_simulation: 1d stimulus array (normalized to [-1; 1]) or [Simulation][remeta.simulation.Simulation] object
        stimuli: 1d stimulus array
        choices: 1d choice array
        params: pass parameters as a dictionary
        type1_noise: Type 1 noise.
        type1_bias: Type 1 bias.
        type1_thresh: Type 1 threshold.
        type2_noise: Metacognitive noise. Required
        type2_evidence_bias_mult: Multiplicative metacognitive bias
        type2_criteria: Confidence criteria
        cfg: Place holder - checking the configuration object is not yet implemented
        stim_max: float | None = None,
        probability_correct: if True, convert confidence (range 0-1) to subjective probability correct (range 0.5-1)
        model_prediction: whether to show model-predicted values for comparison
        model_only: Show the model prediction only (auto-set to True if no data are passed)
        separate_by_category: separate histograms for the two stimulus categories
        separate_by_accuracy: separate histograms for correct and incorrect responses
        model_prediction_nsamples: number of samples used to generate model predictions
    """

    if separate_by_accuracy:
        separate_by_category = False

    if model_only:
        model_prediction = True

    if isinstance(confidence_or_simulation, Simulation):
        from copy import deepcopy
        cfg = deepcopy(confidence_or_simulation.cfg)
        stimuli, confidence, choices = confidence_or_simulation.stimuli, confidence_or_simulation.confidence, confidence_or_simulation.choices
        params = confidence_or_simulation.params.copy()
    else:

        if params is not None or (type1_noise is not None and type2_noise is not None):
            model_prediction = True
        confidence = None if model_only else confidence_or_simulation
        stimuli = None if model_only else stimuli
        choices = None if model_only else choices
        if confidence is None:
            model_only = True

        if separate_by_accuracy and choices is None and not model_only:
            raise ValueError('If separate_by_accuracy is True, choices must be passed.')

        if model_prediction:
            if params is None:
                if type1_noise is None:
                    raise ValueError('Type 1 noise is unspecified.')
                if type2_noise is None:
                    raise ValueError('Type 2 noise is unspecified.')
                params = dict(type1_noise=type1_noise, type2_noise=type2_noise)
                for param in ('type1_bias', 'type1_thresh', 'type2_evidence_bias_mult', 'type2_criteria'):
                    if (value := eval(param)) is not None and not (cfg is not None and not getattr(cfg, f'param_{param}').enable):
                        params[param] = value
            else:
                params = params.copy()
                if 'type1_noise' not in params:
                    raise ValueError('Type 1 noise is unspecified.')
                if 'type2_noise' not in params:
                    raise ValueError('Type 2 noise is unspecified.')
                for param in ('type1_bias', 'type1_thresh', 'type2_evidence_bias_mult', 'type2_criteria'):
                    if param in params and (cfg is not None and not getattr(cfg, f'param_{param}').enable):
                        params.pop(param)


    if probability_correct:
        confidence = (confidence + 1) / 2

    if stim_max is None:
        stim_max = 1 if stimuli is None else np.max(np.abs(stimuli))

    fig = plt.figure(figsize=(6, 3.5))
    fig.subplots_adjust(bottom=0.2)
    gs = fig.add_gridspec(1, 2, width_ratios=[1, 0.5], wspace=0)
    ax = fig.add_subplot(gs[0, 0])
    ax_leg = fig.add_subplot(gs[0, 1])
    ax_leg.axis("off")
    plt.sca(ax)
    ax.xaxis.set_major_formatter(FormatStrFormatter('%.3g'))

    if not model_only:
        conf_levels = np.sort(np.unique(confidence))
        n_conf_levels = len(conf_levels)
        bins = np.linspace(0, 1, n_conf_levels+1) if n_conf_levels >= 8 else np.hstack((0, conf_levels[1:] - np.diff(conf_levels) / 2, 1))

        if separate_by_category:
            counts0 = np.histogram(confidence[np.sign(stimuli) == -1], bins=bins)[0]
            counts1 = np.histogram(confidence[np.sign(stimuli) == 1], bins=bins)[0]
            centers = 0.5 * (bins[:-1] + bins[1:])
            widths  = bins[1:] - bins[:-1]
            for i, (x, w, c0, c1) in enumerate(zip(centers, widths, counts0, counts1)):
                if c0 < c1:
                    plt.bar(x, c1, width=w, color=(0.8, 0.8, 0.8), ec='grey', align='center', zorder=1, label='Data ($S^+$)' if i == 0 else None)
                    plt.bar(x, c0, width=w, color=(0.4, 0.4, 0.4), ec='grey', align='center', zorder=2, label='Data ($S^-$)' if i == 0 else None)
                else:
                    plt.bar(x, c0, width=w, color=(0.4, 0.4, 0.4), ec='grey', align='center', zorder=1, label='Data ($S^-$)' if i == 0 else None)
                    plt.bar(x, c1, width=w, color=(0.8, 0.8, 0.8), ec='grey', align='center', zorder=2, label='Data ($S^+$)' if i == 0 else None)
        elif separate_by_accuracy:
            accuracy = np.sign(stimuli) == np.sign(choices - 0.5)
            counts_inc = np.histogram(confidence[~accuracy], bins=bins)[0]
            counts_cor = np.histogram(confidence[accuracy], bins=bins)[0]
            centers = 0.5 * (bins[:-1] + bins[1:])
            widths  = bins[1:] - bins[:-1]
            for i, (x, w, c0, c1) in enumerate(zip(centers, widths, counts_inc, counts_cor)):
                if c0 < c1:
                    plt.bar(x, c1, width=w, color=color_cor, ec='grey', align='center', zorder=1, label='Data (correct)' if i == 0 else None)
                    plt.bar(x, c0, width=w, color=color_inc, ec='grey', align='center', zorder=2, label='Data (incorrect)' if i == 0 else None)
                else:
                    plt.bar(x, c0, width=w, color=color_inc, ec='grey', align='center', zorder=1, label='Data (incorrect)' if i == 0 else None)
                    plt.bar(x, c1, width=w, color=color_cor, ec='grey', align='center', zorder=2, label='Data (correct)' if i == 0 else None)
        else:
            plt.hist(confidence, bins=bins, color=(0.8, 0.8, 0.8), edgecolor='grey', clip_on=False, label='Data')

    if model_prediction:
        if model_only:
            _nsamples = int(np.ceil(model_prediction_nsamples / 2))
            bins = np.linspace(0, 1, 5)
        else:
            _nsamples = len(confidence)
        stim_min = stim_max / _nsamples
        levels = np.hstack((np.linspace(-stim_max, -stim_min, _nsamples), np.linspace(stim_min, stim_max, _nsamples)))
        # y_decval = stimulus_to_decision_value(levels, params, return_only_decval=True)
        # c_conf = type1_evidence_to_confidence(
        #     z1_type1_evidence=np.abs(y_decval), x_stim=levels,
        #     **params
        # )
        nsubjects = 100 if model_only else 500
        ds = simulate(
            nsubjects=nsubjects,
            params=params, cfg=cfg, custom_stimuli=levels, verbosity=False,
            stim_max=stim_max, squeeze=True, compute_stats=False,
            silence_warnings=True
        )
        c_conf = (ds.confidence + 1) / 2 if probability_correct else ds.confidence

        if separate_by_category:
            counts0_ = [np.histogram(c_conf[s][np.sign(ds.stimuli[s]) == -1], bins=bins)[0] for s in range(nsubjects)]
            counts1_ = [np.histogram(c_conf[s][np.sign(ds.stimuli[s]) == 1], bins=bins)[0] for s in range(nsubjects)]
            counts0 = np.array([(counts0_[s] / (2*_nsamples if model_only else counts0_[s].sum())) * (1 if model_only else (stimuli < 0).sum()) for s in range(nsubjects)])
            counts1 = np.array([(counts1_[s] / (2*_nsamples if model_only else counts1_[s].sum())) * (1 if model_only else (stimuli > 0).sum()) for s in range(nsubjects)])
            plt.errorbar(
                bins[:-1] + np.diff(bins) / 2 - 0.04, counts0.mean(axis=0),
                yerr=np.percentile(counts0, 97.5, axis=0) - np.percentile(counts0, 2.5, axis=0),
                mew=2, elinewidth=2, capsize=7, capthick=2, fmt='o', markersize=8, mec=color_model, mfc=(0.4, 0.4, 0.4),
                color=color_model, lw=2, label='Model prediction ($S^-$)'
            )
            plt.errorbar(
                bins[:-1] + np.diff(bins) / 2 + 0.04, counts1.mean(axis=0),
                yerr=np.percentile(counts1, 97.5, axis=0) - np.percentile(counts1, 2.5, axis=0),
                mew=2, elinewidth=2, capsize=7, capthick=2, fmt='o', markersize=8, mec=color_model, mfc=(0.8, 0.8, 0.8),
                color=color_model, lw=2, label='Model prediction ($S^+$)'
            )
        elif separate_by_accuracy:
            acc = [np.sign(ds.stimuli[s]) == np.sign(ds.choices[s] - 0.5) for s in range(nsubjects)]
            counts_inc_ = [np.histogram(c_conf[s][~acc[s]], bins=bins)[0] for s in range(nsubjects)]
            counts_cor_ = [np.histogram(c_conf[s][acc[s]], bins=bins)[0] for s in range(nsubjects)]
            print(_nsamples, counts_inc_[0].sum()+counts_cor_[0].sum())
            counts_inc = np.array([(counts_inc_[s] / (2*_nsamples if model_only else counts_inc_[s].sum())) * (1 if model_only else (~accuracy).sum())for s in range(nsubjects)])
            counts_cor = np.array([(counts_cor_[s] / (2*_nsamples if model_only else counts_cor_[s].sum())) * (1 if model_only else accuracy.sum()) for s in range(nsubjects)])
            plt.errorbar(
                bins[:-1] + np.diff(bins) / 2 - 0.04, counts_inc.mean(axis=0),
                yerr=np.percentile(counts_inc, 97.5, axis=0) - np.percentile(counts_inc, 2.5, axis=0),
                mew=2, elinewidth=2, capsize=7, capthick=2, fmt='o', markersize=8, mec=color_model, mfc=color_inc,
                color=color_model, lw=2, label='Model prediction\n(incorrect)'
            )
            plt.errorbar(
                bins[:-1] + np.diff(bins) / 2 + 0.04, counts_cor.mean(axis=0),
                yerr=np.percentile(counts_cor, 97.5, axis=0) - np.percentile(counts_cor, 2.5, axis=0),
                mew=2, elinewidth=2, capsize=7, capthick=2, fmt='o', markersize=8, mec=color_model, mfc=color_cor,
                color=color_model, lw=2, label='Model prediction\n(correct)'
            )
        else:
            counts_ = [np.histogram(c_conf[s], bins=bins)[0] for s in range(nsubjects)]
            counts = np.array([(counts_[s] / (2*_nsamples)) * (1 if model_only else len(confidence)) for s in range(nsubjects)])
            plt.errorbar(
                bins[:-1] + np.diff(bins) / 2, counts.mean(axis=0),
                yerr=np.percentile(counts, 97.5, axis=0) - np.percentile(counts, 2.5, axis=0),
                mew=2, elinewidth=2, capsize=7, capthick=2, fmt='o', markersize=8, mec=color_model, mfc=(0.8, 0.8, 0.8),
                color=color_model, lw=2, label='Model prediction'
            )


    plt.xticks(bins)
    plt.xlabel('Confidence')
    plt.ylabel('Probability' if model_only else 'Count')
    plt.xlim(0, 1)

    if model_prediction:
        anot_type2 = []
        cfg = None  # keep in case cfg is implemented in the future
        for i, (k, v) in enumerate(params.items()):
            if k.startswith('type2_'):
                if listlike(v):
                    val = ', '.join([fmp(p) for p in v])
                    anot_type2 += [f"${symbols[k][1:-1]}=" + f"[{val}]$"]
                else:
                    anot_type2 += [f"${symbols[k][1:-1]}={fmp(v)}$"]
        plt.text(1.055, 0.1-0.2*(separate_by_accuracy | separate_by_category), r'Type 2 parameters:' + '\n' + '\n'.join(anot_type2), transform=ax.transAxes,
                 bbox=dict(fc=[1, 1, 1], ec=[0.5, 0.5, 0.5], lw=1, pad=5), fontsize=11)

    # plt.legend(bbox_to_anchor=(1.01, 1), loc="upper left", fontsize=11, handlelength=1)
    ax_leg.legend(*ax.get_legend_handles_labels(), loc="upper left", fontsize=11, handlelength=1)

    set_fontsize(label='default', tick='default')

    if path_export is not None:
        plt.savefig(path_export, bbox_inches='tight', pad_inches=0.02)


def set_fontsize(label=None, xlabel=None, ylabel=None, tick=None, xtick=None, ytick=None, title=None):

    fig = plt.gcf()

    for ax in fig.axes:
        if xlabel is not None:
            ax.xaxis.label.set_size(FONTSIZE['xlabel'] if xlabel == 'default' else xlabel)
        elif label is not None:
            ax.xaxis.label.set_size(FONTSIZE['label'] if label == 'default' else label)
        if ylabel is not None:
            ax.yaxis.label.set_size(FONTSIZE['ylabel'] if ylabel == 'default' else ylabel)
        elif label is not None:
            ax.yaxis.label.set_size(FONTSIZE['label'] if label == 'default' else label)

        if xtick is not None:
            for ticklabel in (ax.get_xticklabels()):
                ticklabel.set_fontsize(FONTSIZE['xtick'] if xtick == 'default' else xtick)
        elif tick is not None:
            for ticklabel in (ax.get_xticklabels()):
                ticklabel.set_fontsize(FONTSIZE['tick'] if tick == 'default' else tick)
        if ytick is not None:
            for ticklabel in (ax.get_yticklabels()):
                ticklabel.set_fontsize(FONTSIZE['ytick'] if ytick == 'default' else ytick)
        elif tick is not None:
            for ticklabel in (ax.get_yticklabels()):
                ticklabel.set_fontsize(FONTSIZE['tick'] if tick == 'default' else tick)

        if title is not None:
            ax.title.set_fontsize(FONTSIZE['title'] if title == 'default' else title)


if __name__ == '__main__':

    import remeta
    np.random.seed(42)  # make notebook reproducible
    import warnings
    warnings.filterwarnings('error')


    # x_stim, d_dec, c_conf, params, y_decval = remeta.load_dataset(
    # 'type2_simple', return_params=True, return_y_decval=True
    # )
    # remeta.plot_evidence_versus_confidence(x_stim, c_conf, y_decval, params, plot_bias_free=True)

    remeta.plot_psychometric(type1_noise=[0.5, 0.2])


# Copyright (C) 2021 Matthias Guggenmos
import os


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


DESCRIPTION = (
    'Reverse engineering of Metacognition toolbox'
)

DISTNAME = 'remeta'
MAINTAINER = 'Matthias Guggenmos'
MAINTAINER_EMAIL = 'mg.corresponding@gmail.com'
VERSION = '0.99.2'
LICENCE = 'MIT License'

INSTALL_REQUIRES = [
    'numpy>=1.20',
    'scipy>=1.3',
    'multiprocessing_on_dill>=3.5.0a4',
    'matplotlib>=3.6.0',
    'numdifftools>=0.9.38']

PACKAGES = ['remeta']

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Programming Language :: Python :: 3.13'
]

try:
    from setuptools import setup

    _has_setuptools = True
except ImportError:
    from distutils.core import setup

if __name__ == '__main__':

    setup(
        name=DISTNAME,
        author=MAINTAINER,
        author_email=MAINTAINER_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        long_description=open('README.md').read(),
        long_description_content_type='text/markdown',
        license=LICENCE,
        version=VERSION,
        install_requires=INSTALL_REQUIRES,
        include_package_data=True,
        packages=PACKAGES,
        classifiers=CLASSIFIERS,
    )



import numpy as np
from scipy.special import expit, ndtr, ndtri

from .util import _check_param, listlike


# def check_criteria(criteria):
#     """
#     Ensure that criteria do not exceed 1
#
#     Parameters
#     ----------
#     criteria : array-like of dtype float
#         Confidence criteria (typically provided as a list)
#
#     Returns
#     ----------
#     criteria : array-like of dtype float
#         Transformed confidence criteria
#     """
#     ind1 = np.where(np.array(criteria) > 1)[0]
#     if len(ind1):
#         raise ValueError
#         for i in ind1[1:]:
#             criteria[i] = 0
#         criteria[ind1[0]] = 1 - np.sum(criteria[:ind1[0]]) + 1e-8
#         # Note: we add 1e-8 to avoid edge cases due to floating point precision
#     return criteria


def logistic(x, type1_noise, squeeze=False):
    """
    Logistic function

    Parameters
    ----------
    x : array-like
        This is typically a stimulus array or a transformed stimulus array (e.g. decision values).
    type1_noise : float
        Type 1 noise parameter.

    Returns
    ----------
    posterior : array-like
        Posterior probability under a logistic model.
    """
    if squeeze:
        # posterior = 1 / (1 + np.exp(-((np.pi / np.sqrt(3)) * x / np.maximum(type1_noise, 0.001)).squeeze(),
        #                             dtype=maxfloat))
        posterior = expit(((np.pi / np.sqrt(3) * x) / np.maximum(type1_noise, 1e-8)).squeeze())
    else:
        # posterior = 1 / (1 + np.exp(-((np.pi / np.sqrt(3)) * x / np.maximum(type1_noise, 0.001)),
        #                             dtype=maxfloat))
        posterior = expit((np.pi / np.sqrt(3) * x) / np.maximum(type1_noise, 1e-8))

    return posterior.astype(np.float64)


def logistic_inv(posterior, type1_noise):
    """
    Inverse logistic function

    Parameters
    ----------
    posterior : array-like
        Posterior probability.
    type1_noise : float
        Type 1 noise parameter.

    Returns
    ----------
    x : array-like
        See docstring of the logistic method.
    """
    x = -(np.sqrt(3) * type1_noise / np.pi) * np.log(1 / posterior - 1)
    return x


def compute_nonlinear_encoding(x_stim, type1_nonlinear_gain=None, type1_nonlinear_scale=None):

    gain = _check_param(0 if type1_nonlinear_gain is None else type1_nonlinear_gain)
    scale = _check_param(1 if type1_nonlinear_scale is None else type1_nonlinear_scale)

    neg, pos = x_stim < 0, x_stim >= 0
    x_stim_nonlinear = np.empty_like(x_stim)
    x_stim_nonlinear[neg] = x_stim[neg] * (1 + gain[0] * (x_stim[neg] / scale[0])**2 /
                                           (1 + (x_stim[neg] / scale[0])**2))
    x_stim_nonlinear[pos] = x_stim[pos] * (1 + gain[1] * (x_stim[pos] / scale[1])**2 /
                                           (1 + (x_stim[pos] / scale[1])**2))

    return x_stim_nonlinear


def compute_signal_dependent_type1_noise(x_stim, type1_noise=None, type1_thresh=None, type1_noise_heteroscedastic=None,
                                         type1_noise_signal_dependency=None, **kwargs):  # noqa
    """
    Compute signal-dependent type 1 noise.

    Parameters
    ----------
    x_stim : array-like
        Array of signed stimulus intensity values, where the sign codes the stimulus category and the absolut value
        codes the intensity. Must be normalized to [-1; 1].
    type1_noise : float or array-like
        Type 1 noise parameter.
    type1_thresh: float or array-like
        Type 1 threshold.
    type1_noise_heteroscedastic : float or array-like
        Signal-dependent type 1 noise parameter.
    type1_noise_signal_dependency : None | str
        Define the signal dependency of type 1 noise. One of None, 'multiplicative', 'power', 'exponential', 'logarithm'.
    kwargs : dict
        Convenience parameter to avoid an error if irrelevant parameters are passed.

    Returns
    ----------
    type1_noise_heteroscedastic : array-like
        Signal-dependent (heteroscedastic) type 1 noise of shape stimuli.shape.
    """

    type1_noise_ = _check_param(type1_noise)
    type1_heteroscedastic_ = _check_param(type1_noise_heteroscedastic)
    type1_thresh_ = (0, 0) if type1_thresh is None else _check_param(type1_thresh)
    type1_noise_heteroscedastic = np.ones(x_stim.shape)
    neg, pos = x_stim < 0, x_stim >= 0
    if type1_noise_signal_dependency is None:
        type1_noise_heteroscedastic[neg] *= type1_noise_[0]
        type1_noise_heteroscedastic[pos] *= type1_noise_[1]
    elif type1_noise_signal_dependency == 'multiplicative':
        # Weber's law with a noise floor (type1_noise) and a noise scaling parameter (type1_heteroscedastic_)
        type1_noise_heteroscedastic[neg] = np.sqrt(type1_noise_[0] ** 2 +
                                                         ((np.abs(x_stim[neg]) - type1_thresh_[0]) * type1_heteroscedastic_[0]) ** 2)  # noqa
        type1_noise_heteroscedastic[pos] = np.sqrt(type1_noise_[1] ** 2 +
                                                         ((np.abs(x_stim[pos]) - type1_thresh_[1]) * type1_heteroscedastic_[1]) ** 2)  # noqa
    elif type1_noise_signal_dependency == 'power':
        type1_noise_heteroscedastic[neg] = np.sqrt(type1_noise_[0] ** 2 +
                                                         (np.abs(x_stim[neg]) - type1_thresh_[0]) ** (2 * type1_heteroscedastic_[0]))  # noqa
        type1_noise_heteroscedastic[pos] = np.sqrt(type1_noise_[1] ** 2 +
                                                         (np.abs(x_stim[pos]) - type1_thresh_[1]) ** (2 * type1_heteroscedastic_[1]))  # noqa
    elif type1_noise_signal_dependency == 'exponential':
        type1_noise_heteroscedastic[neg] = np.sqrt(type1_noise_[0] ** 2 +
                                                         (np.exp(type1_heteroscedastic_[0] * (np.abs(x_stim[neg]) - type1_thresh_[0])) - 1) ** 2)  # noqa
        type1_noise_heteroscedastic[pos] = np.sqrt(type1_noise_[1] ** 2 +
                                                         (np.exp(type1_heteroscedastic_[1] * (np.abs(x_stim[pos]) - type1_thresh_[1])) - 1) ** 2)  # noqa
    elif type1_noise_signal_dependency == 'logarithm':
        type1_noise_heteroscedastic[neg] = np.sqrt(type1_noise_[0] ** 2 +
                                                         np.log(type1_heteroscedastic_[0] * (np.abs(x_stim[neg]) - type1_thresh_[0]) + 1) ** 2)  # noqa
        type1_noise_heteroscedastic[pos] = np.sqrt(type1_noise_[1] ** 2 +
                                                         np.log(type1_heteroscedastic_[1] * (np.abs(x_stim[pos]) - type1_thresh_[1]) + 1) ** 2)  # noqa
    else:
        raise ValueError(f'{type1_noise_signal_dependency} is not a valid function for type1_noise_signal_dependency')

    return type1_noise_heteroscedastic


def type1_evidence_to_confidence(z1_type1_evidence,
                                 type1_dist='normal',
                                 type2_evidence_bias=1,
                                 type2_confidence_bias=1,
                                 type1_noise=None, type1_thresh=None,
                                 type1_noise_heteroscedastic=None, type1_noise_signal_dependency=None,
                                 x_stim=None,
                                 **kwargs):  # noqa
    """
    Transformation from type 1 evidence (z1) to confidence (c).

    Parameters
    ----------
    z1_type1_evidence : array-like
        Evidence at the type 1 level (= absolute decision value).
    type1_dist : str
        Type 1 noise distribution ('normal' or 'logistic')
    type2_evidence_bias : float or array-like
        Multiplicative metacognitive bias parameter loading on evidence.
    type2_confidence_bias : float or array-like
        Power-law confidence bias
    type1_noise : float or array-like
        Type 1 noise parameter. Can be array-like in case of signal-dependent type 1 noise.
    type1_noise_heteroscedastic : float or array-like
        Signal-dependent type 1 noise parameter.
    type1_noise_signal_dependency : str
        Signal-dependent type 1 noise type. One of 'linear', 'power', 'exponential', 'logarithm'.
    x_stim : array-like
        Array of signed stimulus intensity values, where the sign codes the stimulus category and the absolut value
        codes the intensity.
    kwargs : dict
        Convenience parameter to avoid an error if irrelevant parameters are passed.

    Returns
    ----------
    c_conf : array-like
        Model-predicted confidence.
    """
    z1_type1_evidence = np.atleast_1d(z1_type1_evidence)

    if ((type1_noise_signal_dependency is not None) or (listlike(type1_noise) and len(type1_noise) == 2)):
        if x_stim is None:
            raise ValueError('Type 1 noise is signal-dependent, but stimuli (x_stim) have not been '
                             'passed.')
        type1_noise = compute_signal_dependent_type1_noise(
            x_stim.reshape(-1, 1) if (x_stim.ndim == 1) and (z1_type1_evidence.ndim == 2) else x_stim,
            type1_noise=type1_noise, type1_thresh=type1_thresh, type1_noise_heteroscedastic=type1_noise_heteroscedastic,
            type1_noise_signal_dependency=type1_noise_signal_dependency)

    z2_type2_evidence = type2_evidence_bias * z1_type1_evidence

    if type1_dist == 'normal':
        c_conf = (2 * ndtr(z2_type2_evidence / type1_noise) - 1) ** (1/type2_confidence_bias)
    elif type1_dist == 'logistic':
        c_conf = np.tanh(np.pi * z2_type2_evidence / (2 * np.sqrt(3) * type1_noise)) ** (1/type2_confidence_bias)

    return c_conf


def confidence_to_type1_evidence(c_conf,
                                 type1_dist='normal',
                                 type2_evidence_bias=1,
                                 type2_confidence_bias=1,
                                 type1_noise=None, type1_thresh=None,
                                 type1_noise_heteroscedastic=None, type1_noise_signal_dependency=None,
                                 y_decval=None, x_stim=None,
                                 tile_on_type1_uncertainty=True,
                                 **kwargs):  ## noqa
    """
    Transformation from confidence (c) to type 1 evidence (z1).

    Parameters
    ----------
    c_conf : array-like
        Confidence ratings (from behavioral or simulated data).
    type1_dist : str
        Type 1 noise distribution ('normal' or 'logistic')
    type2_evidence_bias : float or array-like
        Multiplicative metacognitive bias parameter loading on evidence.
    type2_confidence_bias : float or array-like
        Power-law confidence bias
    type1_noise : float or array-like
        Type 1 noise parameter. Can be array-like in case of signal-dependent type 1 noise.
    type1_noise_heteroscedastic : float or array-like
        Signal-dependent type 1 noise parameter.
    type1_noise_signal_dependency : str
        Signal-dependent type 1 noise type. One of 'linear', 'power', 'exponential', 'logarithm'.
    y_decval : array-like
        Decision values.
    x_stim : array-like
        Array of signed stimulus intensity values, where the sign codes the stimulus category and the absolut value
        codes the intensity.
    kwargs : dict
        Convenience parameter to avoid an error if irrelevant parameters are passed.

    Returns
    ----------
    z1_type1_evidence : array-like
        Absolute decision values ('z1_type1_evidence').
    """
    if tile_on_type1_uncertainty and y_decval is not None:
        c_conf = np.tile(c_conf, y_decval.shape[-1])

    if ((type1_noise_signal_dependency is not None) or (hasattr(type1_noise, '__len__') and len(type1_noise) == 2)):
        if x_stim is None:
            raise ValueError('Type 1 noise is signal-dependent, but stimuli (x_stim) have not been '
                             'passed.')
        type1_noise = compute_signal_dependent_type1_noise(
            x_stim, type1_noise=type1_noise, type1_thresh=type1_thresh, type1_noise_heteroscedastic=type1_noise_heteroscedastic,
            type1_noise_signal_dependency=type1_noise_signal_dependency)


    if type2_confidence_bias == 1:
        c_conf = np.minimum(1 - 1e-8, c_conf)
    else:
        c_conf =np.minimum(1 - 1e-8, c_conf ** type2_confidence_bias)

    if type1_dist == 'normal':
        z2_type2_evidence = type1_noise * ndtri((c_conf + 1) / 2)
    elif type1_dist == 'logistic':
        z2_type2_evidence = (2 * np.sqrt(3) * type1_noise / np.pi) * np.arctanh(c_conf)

    z1_type1_evidence = z2_type2_evidence / type2_evidence_bias


    return z1_type1_evidence



def confidence_to_type1_noise(c_conf,
                              type1_dist='normal',
                              type2_evidence_bias=1,
                              type2_confidence_bias=1,
                              type1_noise=None, type1_thresh=None,
                              type1_noise_heteroscedastic=None, type1_noise_signal_dependency=None,
                              y_decval=None, x_stim=None,
                              tile_type1_uncertainty=False,
                              **kwargs):  ## noqa
    """
    Transformation from confidence (c) to type 1 evidence (z1).

    Parameters
    ----------
    c_conf : array-like
        Confidence ratings (from behavioral or simulated data).
    type1_dist : str
        Type 1 noise distribution ('normal' or 'logistic')
    type2_evidence_bias : float or array-like
        Multiplicative metacognitive bias parameter loading on evidence.
    type2_confidence_bias : float or array-like
        Power-law confidence bias
    type1_noise : float or array-like
        Type 1 noise parameter. Can be array-like in case of signal-dependent type 1 noise.
    type1_noise_heteroscedastic : float or array-like
        Signal-dependent type 1 noise parameter.
    type1_noise_signal_dependency : str
        Signal-dependent type 1 noise type. One of 'linear', 'power', 'exponential', 'logarithm'.
    y_decval : array-like
        Decision values.
    x_stim : array-like
        Array of signed stimulus intensity values, where the sign codes the stimulus category and the absolut value
        codes the intensity.
    kwargs : dict
        Convenience parameter to avoid an error if irrelevant parameters are passed.

    Returns
    ----------
    z1_type1_evidence : array-like
        Absolute decision values ('z1_type1_evidence').
    """
    if y_decval is None:
        y_decval = c_conf
    elif tile_type1_uncertainty:
        c_conf = np.tile(c_conf, y_decval.shape[-1])

    if ((type1_noise_signal_dependency is not None) or (hasattr(type1_noise, '__len__') and len(type1_noise) == 2)):
        raise ValueError('Signal-dependency or dual parameter mode not yet supported for type2 noise type '
                         '"temperature"')
        # if x_stim is None:
        #     raise ValueError('Type 1 noise is signal-dependent, but stimuli (x_stim) have not been '
        #                      'passed.')
        # type1_noise = compute_signal_dependent_type1_noise(
        #     x_stim, type1_noise=type1_noise, type1_thresh=type1_thresh, type1_noise_heteroscedastic=type1_noise_heteroscedastic,
        #     type1_noise_signal_dependency=type1_noise_signal_dependency)

    if type2_confidence_bias == 1:
        c_conf = np.maximum(1e-8, np.minimum(1 - 1e-8, c_conf))
    else:
        c_conf = np.maximum(1e-8, np.minimum(1 - 1e-8, c_conf ** type2_confidence_bias))

    z2_type2_evidence = np.abs(y_decval) * type2_evidence_bias

    if type1_dist == 'normal':
        type1_noise = z2_type2_evidence / ndtri((c_conf + 1) / 2)
    elif type1_dist == 'logistic':
        type1_noise = (np.pi / (2 * np.sqrt(3))) * (z2_type2_evidence / np.arctanh(c_conf))

    return type1_noise



def type1_noise_to_confidence(type1_noise,
                              type1_dist='normal',
                              type2_evidence_bias=1,
                              type2_confidence_bias=1,
                              type1_thresh=None,
                              type1_noise_heteroscedastic=None, type1_noise_signal_dependency=None,
                              y_decval=None, x_stim=None,
                              **kwargs):  # noqa
    """
    Transformation from type 1 evidence (z1) to confidence (c).

    Parameters
    ----------
    z1_type1_evidence : array-like
        Evidence at the type 1 level (= absolute decision value).
    type1_dist : str
        Type 1 noise distribution ('normal' or 'logistic')
    type2_evidence_bias : float or array-like
        Multiplicative metacognitive bias parameter loading on evidence.
    type2_confidence_bias : float or array-like
        Power-law confidence bias
    type1_noise : float or array-like
        Type 1 noise parameter. Can be array-like in case of signal-dependent type 1 noise.
    type1_noise_heteroscedastic : float or array-like
        Signal-dependent type 1 noise parameter.
    type1_noise_signal_dependency : str
        Signal-dependent type 1 noise type. One of 'linear', 'power', 'exponential', 'logarithm'.
    y_decval : array-like
        Decision values.
    x_stim : array-like
        Array of signed stimulus intensity values, where the sign codes the stimulus category and the absolut value
        codes the intensity.
    kwargs : dict
        Convenience parameter to avoid an error if irrelevant parameters are passed.

    Returns
    ----------
    c_conf : array-like
        Model-predicted confidence.
    """

    if ((type1_noise_signal_dependency is not None) or (hasattr(type1_noise, '__len__') and len(type1_noise) == 2)):
        raise ValueError('Signal-dependency or dual parameter mode not yet supported for type2 noise type '
                         '"temperature"')
        # if x_stim is None:
        #     raise ValueError('Type 1 noise is signal-dependent, but stimuli (x_stim) have not been '
        #                      'passed.')
        # type1_noise = compute_signal_dependent_type1_noise(
        #     x_stim.reshape(-1, 1) if (x_stim.ndim == 1) and (z1_type1_evidence.ndim == 2) else x_stim,
        #     type1_noise=type1_noise, type1_thresh=type1_thresh, type1_noise_heteroscedastic=type1_noise_heteroscedastic,
        #     type1_noise_signal_dependency=type1_noise_signal_dependency)

    z2_type2_evidence = type2_evidence_bias * np.abs(y_decval)

    if type1_dist == 'normal':
        c_conf = (2 * ndtr(z2_type2_evidence / type1_noise) - 1) ** (1/type2_confidence_bias)
    elif type1_dist == 'logistic':
        c_conf = (np.tanh(np.pi * z2_type2_evidence / (2 * np.sqrt(3) * type1_noise))) ** (1/type2_confidence_bias)

    return c_conf
